"""
Module for representing antenna array geometry and for generating
phasing information.
"""
import ephem, math, numpy as np, coord, const, _cephes
from miriad import ij2bl, bl2ij

class PointingError(Exception):
    """An error to throw if a source is below the horizon."""
    def __init__(self, value): self.parameter = value
    def __str__(self): return str(self.parameter)

#  _   _ _   _ _ _ _           _____                 _   _                 
# | | | | |_(_) (_) |_ _   _  |  ___|   _ _ __   ___| |_(_) ___  _ __  ___ 
# | | | | __| | | | __| | | | | |_ | | | | '_ \ / __| __| |/ _ \| '_ \/ __|
# | |_| | |_| | | | |_| |_| | |  _|| |_| | | | | (__| |_| | (_) | | | \__ \
#  \___/ \__|_|_|_|\__|\__, | |_|   \__,_|_| |_|\___|\__|_|\___/|_| |_|___/
#                      |___/        

def juldate2ephem(num):
    """Convert Julian date to ephem date, measured from noon, Dec. 31, 1899."""
    return ephem.date(num - 2415020.)

def ephem2juldate(num):
    """Convert ephem date (measured from noon, Dec. 31, 1899) to Julian date."""
    return float(num + 2415020.)

#  ____           _ _       ____            _       
# |  _ \ __ _  __| (_) ___ | __ )  ___   __| |_   _ 
# | |_) / _` |/ _` | |/ _ \|  _ \ / _ \ / _` | | | |
# |  _ < (_| | (_| | | (_) | |_) | (_) | (_| | |_| |
# |_| \_\__,_|\__,_|_|\___/|____/ \___/ \__,_|\__, |
#                                             |___/ 

class RadioBody:
    """The location of a celestial source."""
    def __init__(self, name, mfreq, ionref, srcshape):
        self.src_name = name
        self.mfreq = mfreq
        self.ionref = list(ionref)
        self.srcshape = list(srcshape)
    def __str__(self):
        return "%s"% (self.src_name,)
    def compute(self, observer):
        """Update coordinates relative to the provided observer.  Must be
        called at each time step before accessing information."""
        # Generate a map for projecting baselines to uvw coordinates
        self.map = coord.eq2top_m(observer.sidereal_time()-self.ra, self.dec)
    def get_crds(self, crdsys, ncrd=3):
        """Return the coordinates of this location in the desired coordinate
        system ('eq','top') in the current epoch.  If ncrd=2, angular
        coordinates (ra/dec or az/alt) are returned, and if ncrd=3,
        xyz coordinates are returned."""
        assert(crdsys in ('eq','top'))
        assert(ncrd in (2,3))
        if crdsys == 'eq':
            if ncrd == 2: return (self.ra, self.dec)
            return coord.radec2eq((self.ra, self.dec))
        else:
            if ncrd == 2: return (self.az, self.alt)
            return coord.azalt2top((self.az, self.alt))

#  ____           _ _       _____ _              _ ____            _       
# |  _ \ __ _  __| (_) ___ |  ___(_)_  _____  __| | __ )  ___   __| |_   _ 
# | |_) / _` |/ _` | |/ _ \| |_  | \ \/ / _ \/ _` |  _ \ / _ \ / _` | | | |
# |  _ < (_| | (_| | | (_) |  _| | |>  <  __/ (_| | |_) | (_) | (_| | |_| |
# |_| \_\__,_|\__,_|_|\___/|_|   |_/_/\_\___|\__,_|____/ \___/ \__,_|\__, |
#                                                                    |___/ 

class RadioFixedBody(ephem.FixedBody, RadioBody):
    """A source at fixed RA,DEC.  Combines ephem.FixedBody with RadioBody."""
    def __init__(self, ra, dec, mfreq=.150, name='', epoch=ephem.J2000,
            ionref=(0.,0.), srcshape=(0.,0.,0.), **kwargs):
        RadioBody.__init__(self, name, mfreq, ionref, srcshape)
        ephem.FixedBody.__init__(self)
        self._ra, self._dec = ra, dec
        self._epoch = epoch
    def __str__(self):
        if self._dec<0: return RadioBody.__str__(self) + '\t' + str(self._ra) +'\t'+ str(self._dec)
        else: return RadioBody.__str__(self) + '\t' + str(self._ra) +'\t'+'+' + str(self._dec)
    def compute(self, observer):
        ephem.FixedBody.compute(self, observer)
        RadioBody.compute(self, observer)

#  ____           _ _      ____                  _       _ 
# |  _ \ __ _  __| (_) ___/ ___| _ __   ___  ___(_) __ _| |
# | |_) / _` |/ _` | |/ _ \___ \| '_ \ / _ \/ __| |/ _` | |
# |  _ < (_| | (_| | | (_) |__) | |_) |  __/ (__| | (_| | |
# |_| \_\__,_|\__,_|_|\___/____/| .__/ \___|\___|_|\__,_|_|
#                               |_|                        

class RadioSpecial(RadioBody, object):
    """A moving source (Sun,Moon,planets).  Combines ephem versions of these 
    objects with RadioBody."""
    def __init__(self,name, mfreq=.150,
            ionref=(0.,0.), srcshape=(0.,0.,0.), **kwargs):
        """`name' is used to lookup appropriate ephem celestial object."""
        RadioBody.__init__(self, name, mfreq, ionref, srcshape)
        self.Body = eval('ephem.%s()' % name)
    def __getattr__(self, nm):
        """First try to access attribute from this class, but if that fails, 
        try to get it from the underlying ephem object."""
        try: return object.__getattr__(self, nm)
        except(AttributeError): return self.Body.__getattribute__(nm)
    def __setattr__(self, nm, val):
        """First try to set attribute for this class, buf if that fails, 
        try to set it for the underlying ephem object."""
        try: object.__setattr__(self, nm, val)
        except(AttributeError): return setattr(self.Body, nm, val)
    def compute(self, observer):
        self.Body.compute(observer)
        RadioBody.compute(self, observer)

#  ____            ____      _        _             
# / ___| _ __ ___ / ___|__ _| |_ __ _| | ___   __ _ 
# \___ \| '__/ __| |   / _` | __/ _` | |/ _ \ / _` |
#  ___) | | | (__| |__| (_| | || (_| | | (_) | (_| |
# |____/|_|  \___|\____\__,_|\__\__,_|_|\___/ \__, |
#                                             |___/ 

class SrcCatalog(dict):
    """A catalog of celestial sources.  Can be initialized with a list
    of src objects, of as an empty catalog."""
    def __init__(self, *srcs, **kwargs):
        dict.__init__(self)
        self.add_srcs(*srcs)
    def add_srcs(self, *srcs):
        """Add src object(s) (RadioFixedBody,RadioSpecial) to catalog."""
        if len(srcs) == 1 and getattr(srcs[0], 'src_name', None) == None:
            srcs = srcs[0]
        for s in srcs: self[s.src_name] = s
    def get_srcs(self, *srcs):
        """Return list of all src objects in catalog."""
        if len(srcs) == 0: srcs = self.keys()
        elif len(srcs) == 1 and type(srcs[0]) != str:
            return [self[s] for s in srcs[0]]
        else: return [self[s] for s in srcs]
    def compute(self, observer):
        """Call compute method of all objects in catalog."""
        for s in self: self[s].compute(observer)
    def get_crds(self, crdsys, ncrd=3, srcs=None):
        """Return coordinates of all objects in catalog."""
        if srcs is None: srcs = self.keys()
        crds = np.array([self[s].get_crds(crdsys, ncrd=ncrd) for s in srcs])
        return crds.transpose()
    def get(self, attribute, srcs=None):
        """Return the specified source attribute (e.g. "mfreq" for src.mfreq)
        in an array for all src names in 'srcs'.  If not provided, defaults to 
        all srcs in catalog."""
        if srcs is None: srcs = self.keys()
        return np.array([getattr(self[s], attribute) for s in srcs]).transpose()

#  ____
# | __ )  ___  __ _ _ __ ___
# |  _ \ / _ \/ _` | '_ ` _ \
# | |_) |  __/ (_| | | | | | |
# |____/ \___|\__,_|_| |_| |_|

class Beam:
    """Template for representing antenna beam pattern.  Beams also hold
    info about which frequencies are active (i.e. Antennas and AntennaArrays
    access frequencies through Beam)."""
    def __init__(self, freqs, **kwargs):
        """freqs = frequencies (in GHz) at bin centers across spectrum."""
        self.freqs = freqs
        self.chans = np.arange(self.freqs.size)
        self._update_afreqs()
    def _update_afreqs(self):
        self.afreqs = self.freqs.take(self.chans)
    def update(self):
        self._update_afreqs()
    def select_chans(self, active_chans=None):
        """Select only enumerated channels to use for future calculations."""
        if active_chans is None: active_chans = np.arange(self.freqs.size)
        self.chans = active_chans
        self.update()

#     _          _                         
#    / \   _ __ | |_ ___ _ __  _ __   __ _ 
#   / _ \ | '_ \| __/ _ \ '_ \| '_ \ / _` |
#  / ___ \| | | | ||  __/ | | | | | | (_| |
# /_/   \_\_| |_|\__\___|_| |_|_| |_|\__,_|

class Antenna:
    """Representation of physical attributes of individual antenna."""
    def __init__(self, x, y, z, beam, phsoff=[0.,0.], **kwargs):
        """x,y,z = antenna coordinates in equatorial (ns) coordinates
        beam = Beam object
        phsoff = polynomial phase vs. frequency.  Phs term that is linear
                 with freq is often called 'delay'."""
        self.pos = np.array((x,y,z), np.float64) # must be float64 for mir
        self.beam = beam
        self._phsoff = phsoff
        self._update_phsoff()
    def select_chans(self, active_chans=None):
        """Select only the specified channels for use in future calculations."""
        self.beam.select_chans(active_chans)
        self.update()
    def _update_phsoff(self):
        self.phsoff = np.polyval(self._phsoff, self.beam.afreqs)
    def update(self):
        self._update_phsoff()
    def __iter__(self): return self.pos.__iter__()
    def __add__(self, a): return self.pos + a.pos
    __radd__ = __add__
    def __neg__(self): return -self.pos
    def __sub__(self, a): return self.pos - a.pos
    def __rsub__(self, a): return a.pos - self.pos

#     _                         _                    _   _             
#    / \   _ __ _ __ __ _ _   _| |    ___   ___ __ _| |_(_) ___  _ __  
#   / _ \ | '__| '__/ _` | | | | |   / _ \ / __/ _` | __| |/ _ \| '_ \ 
#  / ___ \| |  | | | (_| | |_| | |__| (_) | (_| (_| | |_| | (_) | | | |
# /_/   \_\_|  |_|  \__,_|\__, |_____\___/ \___\__,_|\__|_|\___/|_| |_|
#                         |___/                                        

class ArrayLocation(ephem.Observer):
    """The location and time of an observation."""
    def __init__(self, location):
        """location = (lat,long,[elev]) of array"""
        ephem.Observer.__init__(self)
        self.pressure = 0
        if len(location) == 2: self.lat, self.long = location
        else: self.lat, self.long, self.elev = location
        self._update_eq2zen()
    def _update_eq2zen(self):
        self._eq2zen = coord.eq2top_m(0., self.lat)
    def update(self):
        self._update_eq2zen()
    def get_jultime(self):
        """Get current time as a Julian date."""
        return ephem2juldate(self.date)
    def set_jultime(self, t=None):
        """Set current time as a Julian date."""
        if t is None: t = ephem.julian_date()
        self.set_ephemtime(juldate2ephem(t))
    def set_ephemtime(self, t=None):
        """Set current time as derived from the ephem package.  Recalculates
        matrix for projecting baselines into current positions."""
        if t is None: t = ephem.now()
        self.date, self.epoch = t, t
        self._eq2now = coord.rot_m(-self.sidereal_time(), np.array([0.,0.,1.]))

#     _          _                            _                         
#    / \   _ __ | |_ ___ _ __  _ __   __ _   / \   _ __ _ __ __ _ _   _ 
#   / _ \ | '_ \| __/ _ \ '_ \| '_ \ / _` | / _ \ | '__| '__/ _` | | | |
#  / ___ \| | | | ||  __/ | | | | | | (_| |/ ___ \| |  | | | (_| | |_| |
# /_/   \_\_| |_|\__\___|_| |_|_| |_|\__,_/_/   \_\_|  |_|  \__,_|\__, |
#                                                                 |___/ 

class AntennaArray(ArrayLocation):
    """A collection of antennas, their spacings, and location/time of 
    observations."""
    def __init__(self, location, ants, **kwargs):
        """ location = (lat,long,[elev]) of array
        ants = list of Antenna objects."""
        ArrayLocation.__init__(self, location=location)
        self.ants = ants
    def __iter__(self): return self.ants.__iter__()
    def __getitem__(self, *args): return self.ants.__getitem__(*args)
    def __setitem__(self, *args): return self.ants.__setitem__(*args)
    def __len__(self): return self.ants.__len__()
    def update(self):
        ArrayLocation.update(self)
        for a in self: a.update()
    def select_chans(self, active_chans=None):
        """Select which channels are used in computations.  Default is all."""
        for a in self: a.select_chans(active_chans)
        self.update()
    def ij2bl(self, i, j):
        """Convert baseline i,j (0 indexed) to Miriad's (i+1) << 8 | (j+1) 
        indexing scheme."""
        return ij2bl(i,j)
    
    def bl2ij(self, bl):
        """Convert Miriad's (i+1) << 8 | (j+1) baseline indexing scheme to 
        i,j (0 indexed)"""
        return bl2ij(bl)

    def bl_indices(self, auto=True, cross=True):
        """Return bl indices for baselines in the array."""
        if auto:
            if cross: return [self.ij2bl(i,j) 
                for i in range(len(self)) 
                for j in range(i,len(self))]
            else: return [self.ij2bl(i,i) for i in range(len(self))]
        else:
            if cross: return [self.ij2bl(i,j) 
                for i in range(len(self)) 
                for j in range(i+1,len(self))]
            else: return []
    def get_afreqs(self):
        """Return array of frequencies that are active for simulation."""
        return self[0].beam.afreqs
    def get_freqs(self):
        """Return array of (all) frequencies."""
        return self[0].beam.freqs
    def get_baseline(self, i, j, src='z'):
        """Return the baseline corresponding to i,j in various coordinate 
        projections: src='e' for current equatorial, 'z' for zenith 
        topocentric, 'r' for unrotated equatorial, or a RadioBody for
        projection toward that source."""
        bl = self[j] - self[i]
        if type(src) == str:
            if src == 'e': return np.dot(self._eq2now, bl)
            elif src == 'z': return np.dot(self._eq2zen, bl)
            elif src == 'r': return bl
            else: raise ValueError('Unrecognized source:' + src)
        try:
            if src.alt < 0:
                raise PointingError('%s below horizon' % src.src_name)
            m = src.map
        except(AttributeError):
            ra,dec = coord.eq2radec(src)
            m = coord.eq2top_m(self.sidereal_time() - ra, dec)
        return np.dot(m, bl).transpose()
    def get_phs_offset(self, i, j):
        """Return the frequency-dependent phase offset of baseline i,j."""
        return self[j].phsoff - self[i].phsoff
    def gen_uvw(self, i, j, src='z', w_only=False):
        """Compute uvw coordinates of baseline relative to provided RadioBody, 
        or 'z' for zenith uvw coordinates.  If w_only is True, only w (instead
        of (u,v,w) will be returned)."""
        x,y,z = self.get_baseline(i,j, src=src)
        afreqs = self.get_afreqs()
        afreqs = np.reshape(afreqs, (1,afreqs.size))
        if len(x.shape) == 0:
            if w_only: return z*afreqs
            else: return np.array([x*afreqs, y*afreqs, z*afreqs])
        #afreqs = np.reshape(afreqs, (1,afreqs.size))
        x.shape += (1,); y.shape += (1,); z.shape += (1,)
        if w_only: return np.dot(z,afreqs)
        else: return np.array([np.dot(x,afreqs), np.dot(y,afreqs), np.dot(z,afreqs)])
    def gen_phs(self, src, i, j, mfreq=.150, ionref=None, srcshape=None, 
            resolve_src=False):
        """Return phasing that is multiplied to data to point to src."""
        if ionref is None:
            try: ionref = src.ionref
            except(AttributeError): pass
        if not ionref is None or resolve_src: u,v,w = self.gen_uvw(i,j,src=src)
        else: w = self.gen_uvw(i,j,src=src, w_only=True)
        if not ionref is None: w += self.refract(u, v, mfreq=mfreq, ionref=ionref)
        o = self.get_phs_offset(i,j)
        phs = np.exp(-1j*2*np.pi*(w + o))
        if resolve_src:
            if srcshape is None:
                try: srcshape = src.srcshape
                except(AttributeError): pass
            if not srcshape is None: phs *= self.resolve_src(u, v, srcshape=srcshape)
        return phs.squeeze()
    def resolve_src(self, u, v, srcshape=(0,0,0)):
        """Adjust amplitudes to reflect resolution effects for a uniform 
        elliptical disk characterized by srcshape:
        srcshape = (a1,a2,th) where a1,a2 are angular sizes along the 
            semimajor, semiminor axes, and th is the angle (in radians) of
            the semimajor axis from E."""
        a1,a2,th = srcshape
        try:
            if len(u.shape) > len(a1.shape): 
                a1.shape += (1,); a2.shape += (1,); th.shape += (1,)
        except(AttributeError): pass
        ru = a1 * (u*np.cos(th) - v*np.sin(th))
        rv = a2 * (u*np.sin(th) + v*np.cos(th))
        x = 2 * np.pi * np.sqrt(ru**2 + rv**2)
        # Use first Bessel function of the first kind (J_1)
        return np.where(x == 0, 1, 2 * _cephes.j1(x)/x).squeeze()
    def refract(self, u_sf, v_sf, mfreq=.150, ionref=(0.,0.)):
        """Calibrate a frequency-dependent source offset by scaling measured
        offsets at a given frequency.  Generates dw, a change in the
        projection of a baseline towards that source, which can be used to
        fix the computed phase of that source.
        ionref = (dra, ddec) where dra, ddec are angle offsets (in radians)
            of sources along ra/dec axes at the specified mfreq.
        u_sf,v_sf = u,v components of baseline, used to compute the
            change in w given angle offsets and the small angle approx.  Should
            be numpy arrays with sources (s) along the 1st axis and
            freqs (f) along the 2nd."""
        dra,ddec = ionref
        s,f = u_sf.shape
        try: dra.shape = (s,1)
        except(AttributeError): pass
        try: ddec.shape = (s,1)
        except(AttributeError): pass
        try: mfreq.shape = (s,1)
        except(AttributeError): pass
        f2 = self.get_afreqs()**2 ; f2.shape = (1, f)
        return (dra*u_sf + ddec*v_sf) * mfreq**2 / f2
    def phs2src(self, data, src, i, j, mfreq=.150, ionref=None, srcshape=None):
        """Apply phasing to zenith-phased data to point to src."""
        return data * self.gen_phs(src, i, j, 
            mfreq=mfreq, ionref=ionref, srcshape=srcshape, resolve_src=False)
    def unphs2src(self,data,src, i, j, mfreq=.150, ionref=None, srcshape=None):
        """Remove phasing from src-phased data to point to zenith."""
        return data / self.gen_phs(src, i, j,
            mfreq=mfreq, ionref=ionref, srcshape=srcshape, resolve_src=False)



"""
Nonlinear solvers
=================

These solvers find x for which F(x)=0. Both x and F is multidimensional.

They accept the user defined function F, which accepts a python tuple x and it
should return F(x), which can be either a tuple, or numpy array.

Example:

def F(x):
    "Should converge to x=[0,0,0,0,0]"
    import numpy
    d = numpy.array([3,2,1.5,1,0.5])
    c = 0.01
    return -d*numpy.array(x)-c*numpy.array(x)**3

from aipy import optimize
x = optimize.broyden2(F,[1,1,1,1,1])

All solvers have the parameter iter (the number of iterations to compute), some
of them have other parameters of the solver, see the particular solver for
details.

 A collection of general-purpose nonlinear multidimensional solvers.

   broyden1            --  Broyden's first method - is a quasi-Newton-Raphson
                           method for updating an approximate Jacobian and then
                           inverting it
   broyden2            --  Broyden's second method - the same as broyden1, but
                           updates the inverse Jacobian directly
   broyden3            --  Broyden's second method - the same as broyden2, but
                           instead of directly computing the inverse Jacobian,
                           it remembers how to construct it using vectors, and
                           when computing inv(J)*F, it uses those vectors to
                           compute this product, thus avoding the expensive NxN
                           matrix multiplication.  
   broyden_generalized --  Generalized Broyden's method, the same as broyden2,
                           but instead of approximating the full NxN Jacobian,
                           it construct it at every iteration in a way that
                           avoids the NxN matrix multiplication.  This is not
                           as precise as broyden3.
   anderson            --  extended Anderson method, the same as the
                           broyden_generalized, but added w_0^2*I to before
                           taking inversion to improve the stability
   anderson2           --  the Anderson method, the same as anderson, but
                           formulated differently

 The broyden2 is the best. For large systems, use broyden3. excitingmixing is
 also very effective. There are some more solvers implemented (see their
 docstrings), however, those are of mediocre quality.


 Utility Functions

   norm --  Returns an L2 norm of the vector

"""

import math

import numpy

def mlog(x):
    if x==0.:
        return 13
    else:
        return math.log(x)

def norm(v):
    """Returns an L2 norm of the vector."""
    return math.sqrt(numpy.sum((numpy.array(v)**2).flat))

def myF(F,xm):
    return numpy.matrix(F(tuple(xm.flat))).T

def difference(a,b):
    m=0.
    for x,y in zip(a,b):
        m+=(x-y)**2
    return math.sqrt(m)

def sum(a,b):
    return [ai+bi for ai,bi in zip(a,b)]

def mul(C,b):
    return [C*bi for bi in b]

def solve(A,b):
    """Solve Ax=b, returns x"""
    try:
        #from scipy import linalg
        from numpy import linalg
        return linalg.solve(A,b)
    except:
        return A.I*b

def broyden2(F, xin, iter=10, alpha=0.4, verbose = False):
    """Broyden's second method.

    Updates inverse Jacobian by an optimal formula.
    There is NxN matrix multiplication in every iteration.

    The best norm |F(x)|=0.003 achieved in ~20 iterations.

    Recommended.
    """
    xm=numpy.matrix(xin).T
    Fxm=myF(F,xm)
    Gm=-alpha*numpy.matrix(numpy.identity(len(xin)))
    for n in range(iter):
        deltaxm=-Gm*Fxm
        xm=xm+deltaxm
        Fxm1=myF(F,xm)
        deltaFxm=Fxm1-Fxm
        Fxm=Fxm1
        Gm=Gm+(deltaxm-Gm*deltaFxm)*deltaFxm.T/norm(deltaFxm)**2
        if verbose:
            print "%d:  |F(x)|=%.3f"%(n+1, norm(Fxm))
    return xm.flat

def broyden3(F, xin, iter=10, alpha=0.4, verbose = False):
    """Broyden's second method.

    Updates inverse Jacobian by an optimal formula.
    The NxN matrix multiplication is avoided. 

    The best norm |F(x)|=0.003 achieved in ~20 iterations.

    Recommended.
    """
    zy=[]
    def updateG(z,y):
        "G:=G+z*y.T"
        zy.append((z,y))
    def Gmul(f):
        "G=-alpha*1+z*y.T+z*y.T ..."
        s=-alpha*f
        for z,y in zy:
            s=s+z*(y.T*f)
        return s
    xm=numpy.matrix(xin).T
    Fxm=myF(F,xm)
#    Gm=-alpha*numpy.matrix(numpy.identity(len(xin)))
    for n in range(iter):
        #deltaxm=-Gm*Fxm
        deltaxm=Gmul(-Fxm)
        xm=xm+deltaxm
        Fxm1=myF(F,xm)
        deltaFxm=Fxm1-Fxm
        Fxm=Fxm1
        #Gm=Gm+(deltaxm-Gm*deltaFxm)*deltaFxm.T/norm(deltaFxm)**2
        updateG(deltaxm-Gmul(deltaFxm),deltaFxm/norm(deltaFxm)**2)
        if verbose:
            print "%d:  |F(x)|=%.3f"%(n+1, norm(Fxm))
    return xm.flat

def broyden_generalized(F, xin, iter=10, alpha=0.1, M=5, verbose = False):
    """Generalized Broyden's method.

    Computes an approximation to the inverse Jacobian from the last M
    interations. Avoids NxN matrix multiplication, it only has MxM matrix
    multiplication and inversion.

    M=0 .... linear mixing
    M=1 .... Anderson mixing with 2 iterations
    M=2 .... Anderson mixing with 3 iterations
    etc.
    optimal is M=5

    """
    xm=numpy.matrix(xin).T
    Fxm=myF(F,xm)
    G0=-alpha
    dxm=[]
    dFxm=[]
    for n in range(iter):
        deltaxm=-G0*Fxm
        if M>0:
            MM=min(M,n)
            for m in range(n-MM,n):
                deltaxm=deltaxm-(float(gamma[m-(n-MM)])*dxm[m]-G0*dFxm[m])
        xm=xm+deltaxm
        Fxm1=myF(F,xm)
        deltaFxm=Fxm1-Fxm
        Fxm=Fxm1

        if M>0:
            dxm.append(deltaxm)
            dFxm.append(deltaFxm)
            MM=min(M,n+1)
            a=numpy.matrix(numpy.empty((MM,MM)))
            for i in range(n+1-MM,n+1):
                for j in range(n+1-MM,n+1):
                    a[i-(n+1-MM),j-(n+1-MM)]=dFxm[i].T*dFxm[j]

            dFF=numpy.matrix(numpy.empty(MM)).T
            for k in range(n+1-MM,n+1):
                dFF[k-(n+1-MM)]=dFxm[k].T*Fxm
            gamma=a.I*dFF

        if verbose:
            print "%d:  |F(x)|=%.3f"%(n, norm(Fxm))
    return xm.flat

def anderson(F, xin, iter=10, alpha=0.1, M=5, w0=0.01, verbose = False):
    """Extended Anderson method.

    Computes an approximation to the inverse Jacobian from the last M
    interations. Avoids NxN matrix multiplication, it only has MxM matrix
    multiplication and inversion.

    M=0 .... linear mixing
    M=1 .... Anderson mixing with 2 iterations
    M=2 .... Anderson mixing with 3 iterations
    etc.
    optimal is M=5

    """
    xm=numpy.matrix(xin).T
    Fxm=myF(F,xm)
    dxm=[]
    dFxm=[]
    for n in range(iter):
        deltaxm=alpha*Fxm
        if M>0:
            MM=min(M,n)
            for m in range(n-MM,n):
                deltaxm=deltaxm-(float(gamma[m-(n-MM)])*dxm[m]+alpha*dFxm[m])
        xm=xm+deltaxm
        Fxm1=myF(F,xm)
        deltaFxm=Fxm1-Fxm
        Fxm=Fxm1

        if M>0:
            dxm.append(deltaxm)
            dFxm.append(deltaFxm)
            MM=min(M,n+1)
            a=numpy.matrix(numpy.empty((MM,MM)))
            for i in range(n+1-MM,n+1):
                for j in range(n+1-MM,n+1):
                    if i==j: wd=w0**2
                    else: wd=0
                    a[i-(n+1-MM),j-(n+1-MM)]=(1+wd)*dFxm[i].T*dFxm[j]

            dFF=numpy.matrix(numpy.empty(MM)).T
            for k in range(n+1-MM,n+1):
                dFF[k-(n+1-MM)]=dFxm[k].T*Fxm
            gamma=solve(a,dFF)
#            print gamma

        if verbose:
            print "%d:  |F(x)|=%.3f"%(n, norm(Fxm))
    return xm.flat

def anderson2(F, xin, iter=10, alpha=0.1, M=5, w0=0.01, verbose = False):
    """Anderson method.

    M=0 .... linear mixing
    M=1 .... Anderson mixing with 2 iterations
    M=2 .... Anderson mixing with 3 iterations
    etc.
    optimal is M=5

    """
    xm=numpy.matrix(xin).T
    Fxm=myF(F,xm)
    dFxm=[]
    for n in range(iter):
        deltaxm=Fxm
        if M>0:
            MM=min(M,n)
            for m in range(n-MM,n):
                deltaxm=deltaxm+float(theta[m-(n-MM)])*(dFxm[m]-Fxm)
        deltaxm=deltaxm*alpha
        xm=xm+deltaxm
        Fxm1=myF(F,xm)
        deltaFxm=Fxm1-Fxm
        Fxm=Fxm1

        if M>0:
            dFxm.append(Fxm-deltaFxm)
            MM=min(M,n+1)
            a=numpy.matrix(numpy.empty((MM,MM)))
            for i in range(n+1-MM,n+1):
                for j in range(n+1-MM,n+1):
                    if i==j: wd=w0**2
                    else: wd=0
                    a[i-(n+1-MM),j-(n+1-MM)]= \
                        (1+wd)*(Fxm-dFxm[i]).T*(Fxm-dFxm[j])

            dFF=numpy.matrix(numpy.empty(MM)).T
            for k in range(n+1-MM,n+1):
                dFF[k-(n+1-MM)]=(Fxm-dFxm[k]).T*Fxm
            theta=solve(a,dFF)
#            print gamma

        if verbose:
            print "%d:  |F(x)|=%.3f"%(n, norm(Fxm))
    return xm.flat

def broyden_modified(F, xin, iter=10, alpha=0.35, w0=0.01, wl=5, verbose = False):
    """Modified Broyden's method.

    Updates inverse Jacobian using information from all the iterations and
    avoiding the NxN matrix multiplication. The problem is with the weights,
    it converges the same or worse than broyden2 or broyden_generalized

    """
    xm=numpy.matrix(xin).T
    Fxm=myF(F,xm)
    G0=alpha
    w=[]
    u=[]
    dFxm=[]
    for n in range(iter):
        deltaxm=G0*Fxm
        for i in range(n):
            for j in range(n):
                deltaxm-=w[i]*w[j]*betta[i,j]*u[j]*(dFxm[i].T*Fxm)
        xm+=deltaxm
        Fxm1=myF(F,xm)
        deltaFxm=Fxm1-Fxm
        Fxm=Fxm1

        w.append(wl/norm(Fxm))

        u.append((G0*deltaFxm+deltaxm)/norm(deltaFxm))
        dFxm.append(deltaFxm/norm(deltaFxm))
        a=numpy.matrix(numpy.empty((n+1,n+1)))
        for i in range(n+1):
            for j in range(n+1):
                a[i,j]=w[i]*w[j]*dFxm[j].T*dFxm[i]
        betta=(w0**2*numpy.matrix(numpy.identity(n+1))+a).I

        if verbose:
            print "%d:  |F(x)|=%.3f"%(n, norm(Fxm))
    return xm.flat

def broyden1(F, xin, iter=10, alpha=0.1, verbose = False):
    """Broyden's first method.

    Updates Jacobian and computes inv(J) by a matrix inversion at every
    iteration. It's very slow.

    The best norm |F(x)|=0.005 achieved in ~45 iterations.
    """
    xm=numpy.matrix(xin).T
    Fxm=myF(F,xm)
    Jm=-1/alpha*numpy.matrix(numpy.identity(len(xin)))

    for n in range(iter):
        deltaxm=solve(-Jm,Fxm)
        #!!!! What the fuck?!
        #xm+=deltaxm
        xm=xm+deltaxm
        Fxm1=myF(F,xm)
        deltaFxm=Fxm1-Fxm
        Fxm=Fxm1
        Jm=Jm+(deltaFxm-Jm*deltaxm)*deltaxm.T/norm(deltaxm)**2
        if verbose:
            print "%d:  |F(x)|=%.3f"%(n, norm(Fxm))
    return xm.flat

def broyden1_modified(F, xin, iter=10, alpha=0.1, verbose = False):
    """Broyden's first method, modified by O. Certik.

    Updates inverse Jacobian using some matrix identities at every iteration,
    its faster then newton_slow, but still not optimal.

    The best norm |F(x)|=0.005 achieved in ~45 iterations.
    """
    def inv(A,u,v):

        #interesting is that this 
        #return (A.I+u*v.T).I
        #is more stable than
        #return A-A*u*v.T*A/float(1+v.T*A*u)
        Au=A*u
        return A-Au*(v.T*A)/float(1+v.T*Au)
    xm=numpy.matrix(xin).T
    Fxm=myF(F,xm)
    Jm=alpha*numpy.matrix(numpy.identity(len(xin)))
    for n in range(iter):
        deltaxm=Jm*Fxm
        xm=xm+deltaxm
        Fxm1=myF(F,xm)
        deltaFxm=Fxm1-Fxm
        Fxm=Fxm1
#        print "-------------",norm(deltaFxm),norm(deltaxm)
        deltaFxm/=norm(deltaxm)
        deltaxm/=norm(deltaxm)
        Jm=inv(Jm+deltaxm*deltaxm.T*Jm,-deltaFxm,deltaxm)
        
        if verbose:
            print "%d:  |F(x)|=%.3f"%(n, norm(Fxm))
    return xm

def vackar(F, xin, iter=10, alpha=0.1, verbose = False):
    """J=diag(d1,d2,...,dN)

    The best norm |F(x)|=0.005 achieved in ~110 iterations.
    """
    def myF(F,xm):
        return numpy.array(F(tuple(xm.flat))).T
    xm=numpy.array(xin)
    Fxm=myF(F,xm)
    d=1/alpha*numpy.ones(len(xin))
    Jm=numpy.matrix(numpy.diag(d))

    for n in range(iter):
        deltaxm=1/d*Fxm
        xm=xm+deltaxm
        Fxm1=myF(F,xm)
        deltaFxm=Fxm1-Fxm
        Fxm=Fxm1
        d=d-(deltaFxm+d*deltaxm)*deltaxm/norm(deltaxm)**2
        if verbose:
            print "%d:  |F(x)|=%.3f"%(n, norm(Fxm))
    return xm

def linearmixing(F,xin, iter=10, alpha=0.1, verbose = False):
    """J=-1/alpha
    
    The best norm |F(x)|=0.005 achieved in ~140 iterations.
    """
    def myF(F,xm):
        return numpy.array(F(tuple(xm.flat))).T
    xm=numpy.array(xin)
    Fxm=myF(F,xm)
    for n in range(iter):
        deltaxm=alpha*Fxm
        xm=xm+deltaxm
        Fxm1=myF(F,xm)
        deltaFxm=Fxm1-Fxm
        Fxm=Fxm1
        if verbose:
            print "%d: |F(x)|=%.3f" %(n,norm(Fxm))

    return xm

def excitingmixing(F,xin,iter=10,alpha=0.1,alphamax=1.0, verbose = False):
    """J=-1/alpha
    
    The best norm |F(x)|=0.005 achieved in ~140 iterations.
    """
    def myF(F,xm):
        return numpy.array(F(tuple(xm.flat))).T
    xm=numpy.array(xin)
    beta=numpy.array([alpha]*len(xm))
    Fxm=myF(F,xm)
    for n in range(iter):
        deltaxm=beta*Fxm
        xm=xm+deltaxm
        Fxm1=myF(F,xm)
        deltaFxm=Fxm1-Fxm
        for i in range(len(xm)):
            if Fxm1[i]*Fxm[i] > 0:
                beta[i]=beta[i]+alpha
                if beta[i] > alphamax:
                    beta[i] = alphamax
            else:
                beta[i]=alpha
        Fxm=Fxm1
        if verbose:
            print "%d: |F(x)|=%.3f" %(n,norm(Fxm))

    return xm



import aipy as a

def get_aa(freqs):
    # Define the location of your instrument
    lat, lon = '45:00', '90:00'
    # Create a model of the primary beam.  BeamFlat is a minimal model 
    # that has unity gain in all directions.  
    beam = a.fit.Beam(freqs)
    # Make a list of antennas with requisite nanosecond locations, 
    # primary beams, and any other calibration parameters you wish to provide.
    ants = [
        a.fit.Antenna(  0,   0,  0, beam),
        a.fit.Antenna(  0, 100,  0, beam),
        a.fit.Antenna(100,   0,  0, beam),
        a.fit.Antenna(100, 100,  0, beam),
    ]
    # Create an AntennaArray at the specified location with the listed antennas
    aa = a.fit.AntennaArray((lat,lon), ants)
    return aa

def get_catalog(srcs=None, cutoff=None, catalogs=[]):
    # Pass off the request for sources to the AIPY source catalog.  If desired, 
    # you can substitute your own sources or source calibrations.
    return a.src.get_catalog(srcs=srcs, cutoff=cutoff, catalogs=catalogs)




# -*- coding: utf-8 -*-
import unittest, ephem as e, random
import aipy as a, numpy as n

class TestConvert(unittest.TestCase):
    def testtypecheck(self):
        """Test coordinate conversion types"""
        crd2, crd3 = (0, 0), (1, 0, 0)
        for sys in a.coord.sys_dict:
            ans2 = a.coord.convert(crd2, sys, sys)
            self.assertEqual(crd2[0], ans2[0])
            self.assertEqual(crd2[1], ans2[1])
            ans3 = a.coord.convert(crd3, sys, sys)
            self.assertEqual(crd2[0], ans3[0])
            self.assertEqual(crd2[1], ans3[1])
        self.assertRaises(KeyError, a.coord.convert, crd2, 'bad','bad')
    def testprecession(self):
        """Test coordinate precessions for accuracy"""
        crdpairs = [[('0:00','0:00'), ('00:02:33.77','00:16:42.1')],
            [('6:00','0:00'), ('06:02:33.75','-00:00:05.6')],
            [('0:00','89:00'), ('00:03:03.75','+89:16:41.7')]]
        for b1950,j2000 in crdpairs:
            c1 = a.coord.convert(b1950,'eq','eq',iepoch=e.B1950,oepoch=e.J2000)
            c2 = a.coord.convert(j2000,'eq','eq',iepoch=e.J2000,oepoch=e.B1950)
            c1_ck = e.Equatorial(j2000[0],j2000[1],epoch=e.J2000).get()
            c2_ck = e.Equatorial(b1950[0],b1950[1],epoch=e.B1950).get()
            self.assertAlmostEqual(c1[0], c1_ck[0], 4)
            self.assertAlmostEqual(c1[1], c1_ck[1], 4)
            self.assertAlmostEqual(c2[0], c2_ck[0], 4)
            self.assertAlmostEqual(c2[1], c2_ck[1], 4)
    def testcrdsys(self):
        """Test coordinate conversions for accuracy"""
        eq_ec_ga = [
            [   ('19:59:28.3566','40:44:02.096'),
                ('317.7054323','59.3254895'),
                ('76.1898379','5.7554756')],
            [   ('12:30:49.4233','12:23:28.043'),
                ('182.0592608','14.4166861'),
                ('283.7777978','74.4911308')],
            [   ('13:25:27.6152','-43:01:08.805'),
                ('217.1433477','-31.3319020'),
                ('309.5158743','19.4173247')]]
        for eq,ec,ga in eq_ec_ga:
            eq_ec = a.coord.convert(eq,'eq','ec')
            eq_ga = a.coord.convert(eq,'eq','ga')
            ec_eq = a.coord.convert(ec,'ec','eq')
            ec_ga = a.coord.convert(ec,'ec','ga')
            ga_eq = a.coord.convert(ga,'ga','eq')
            ga_ec = a.coord.convert(ga,'ga','ec')
            eq_ck = e.Equatorial(eq[0], eq[1], epoch=e.J2000).get()
            ec_ck = e.Ecliptic(ec[0], ec[1], epoch=e.J2000).get()
            ga_ck = e.Galactic(ga[0], ga[1], epoch=e.J2000).get()
            self.assertAlmostEqual(eq_ec[0], ec_ck[0], 4)
            self.assertAlmostEqual(eq_ec[1], ec_ck[1], 4)
            self.assertAlmostEqual(eq_ga[0], ga_ck[0], 4)
            self.assertAlmostEqual(eq_ga[1], ga_ck[1], 4)
            self.assertAlmostEqual(ec_eq[0], eq_ck[0], 4)
            self.assertAlmostEqual(ec_eq[1], eq_ck[1], 4)
            self.assertAlmostEqual(ec_ga[0], ga_ck[0], 4)
            self.assertAlmostEqual(ec_ga[1], ga_ck[1], 4)
            self.assertAlmostEqual(ga_eq[0], eq_ck[0], 4)
            self.assertAlmostEqual(ga_eq[1], eq_ck[1], 4)
            self.assertAlmostEqual(ga_ec[0], ec_ck[0], 4)
            self.assertAlmostEqual(ga_ec[1], ec_ck[1], 4)

class TestConvert_m(unittest.TestCase):
    def testshape(self):
        """Test conversion matrices for shape preservation"""
        for c1 in a.coord.sys_dict:
            for c2 in a.coord.sys_dict:
                m = a.coord.convert_m(c1,c2)
                self.assertEqual(len(m.shape), 2)
                self.assertEqual(m.shape[0], 3)
                self.assertEqual(m.shape[1], 3)
    def testdiagonal(self):
        """Test conversion matrices for normalcy"""
        diag = n.array([[1,0,0],[0,1,0],[0,0,1]], dtype=n.double)
        for c1 in a.coord.sys_dict:
            for c2 in a.coord.sys_dict:
                m1 = a.coord.convert_m(c1,c2)
                m2 = a.coord.convert_m(c2,c1)
                m = n.round(n.dot(m1, m2), 10)
                self.assertTrue(n.all(m == diag))

class TestRot_m(unittest.TestCase):
    def testshape(self):
        """Test rotation matrices for shape preservation"""
        m = a.coord.rot_m(0, n.array([1.,0,0]))
        self.assertEqual(len(m.shape), 2)
        self.assertEqual(m.shape[0], 3)
        self.assertEqual(m.shape[1], 3)
        m = a.coord.rot_m(n.array([0., 0.]), n.array([[1.,0,0],[1.,0,0]]))
        self.assertEqual(len(m.shape), 3)
        self.assertEqual(m.shape[0], 2)
        self.assertEqual(m.shape[1], 3)
        self.assertEqual(m.shape[2], 3)
    def testaccuracy(self):
        """Test rotation matricies for accuracy"""
        diag = n.array([[1,0,0],[0,1,0],[0,0,1]], dtype=n.double)
        e1,e2,e3 = diag
        m = n.round(a.coord.rot_m(2*n.pi, n.array([1.,0,0])), 10)
        self.assertTrue(n.all(m == diag))
        m = n.round(a.coord.rot_m(n.pi/2, n.array([0.,0,1])), 10)
        self.assertTrue(n.all(m == n.array([-e2,e1,e3])))
        m = n.round(a.coord.rot_m(n.pi/2, n.array([0,1,0])), 10)
        self.assertTrue(n.all(m == n.array([e3,e2,-e1])))
        m = n.round(a.coord.rot_m(n.pi/2, n.array([1,0,0])), 10)
        self.assertTrue(n.all(m == n.array([e1,-e3,e2])))
        
class TestXyz2thphi(unittest.TestCase):
    def testshape(self):
        """Test the x,y,z to theta,phi conversion for shape preservation"""
        m = a.coord.xyz2thphi((0,0,1))
        self.assertEqual(len(m.shape), 1)
        self.assertEqual(m.shape[0], 2)
        m = a.coord.xyz2thphi((n.array([0.,0,0]),
            n.array([0.,0,0]),n.array([1.,1,1])))
        self.assertEqual(len(m.shape), 2)
        self.assertEqual(m.shape[0], 2)
        self.assertEqual(m.shape[1], 3)
    def testaccuracy(self):
        """Test the x,y,z to theta,phi conversion for accuracy"""
        diag = n.array([[1,0,0],[0,1,0],[0,0,1]], dtype=n.double)
        e1,e2,e3 = diag
        m = n.round(a.coord.xyz2thphi(e3), 10)
        self.assertTrue(n.all(m == n.array([0.,0])))
        m = n.round(a.coord.xyz2thphi(e1), 10)
        self.assertTrue(n.all(m == n.round(n.array([n.pi/2,0]), 10)))
        m = n.round(a.coord.xyz2thphi(e2), 10)
        self.assertTrue(n.all(m == n.round(n.array([n.pi/2,n.pi/2]), 10)))
        
class TestThphi2xyz(unittest.TestCase):
    def testshape(self):
        """Test the theta,phi to x,y,z conversion for shape preservation"""
        m = a.coord.thphi2xyz((0,0))
        self.assertEqual(len(m.shape), 1)
        self.assertEqual(m.shape[0], 3)
        m = a.coord.thphi2xyz((n.array([0.,0]),n.array([0.,0])))
        self.assertEqual(len(m.shape), 2)
        self.assertEqual(m.shape[0], 3)
        self.assertEqual(m.shape[1], 2)
    def testaccuracy(self):
        """Test the theta,phi to x,y,z conversion for accuracy"""
        diag = n.array([[1,0,0],[0,1,0],[0,0,1]], dtype=n.double)
        e1,e2,e3 = diag
        m = n.round(a.coord.thphi2xyz((0,0)), 10)
        self.assertTrue(n.all(m == e3))
        m = n.round(a.coord.thphi2xyz((n.pi/2,0)), 10)
        self.assertTrue(n.all(m == e1))
        m = n.round(a.coord.thphi2xyz((n.pi/2,n.pi/2)), 10)
        self.assertTrue(n.all(m == e2))
        
class TestEq2radec(unittest.TestCase):
    def testshape(self):
        """Test the equatorial to ra,dec conversion for shape preservation"""
        m = a.coord.eq2radec((0,0,1))
        self.assertEqual(len(m.shape), 1)
        self.assertEqual(m.shape[0], 2)
        m = a.coord.eq2radec((n.array([0.,0,0]),
            n.array([0.,0,0]),n.array([1.,1,1])))
        self.assertEqual(len(m.shape), 2)
        self.assertEqual(m.shape[0], 2)
        self.assertEqual(m.shape[1], 3)
    def testaccuracy(self):
        """Test the equatorial to ra,dec conversion for accuracy"""
        diag = n.array([[1,0,0],[0,1,0],[0,0,1]], dtype=n.double)
        e1,e2,e3 = diag
        m = n.round(a.coord.eq2radec(e3), 10)
        self.assertTrue(n.all(m == n.round(n.array([0.,n.pi/2]), 10)))
        m = n.round(a.coord.eq2radec(e1), 10)
        self.assertTrue(n.all(m == n.round(n.array([0,0]), 10)))
        m = n.round(a.coord.eq2radec(e2), 10)
        self.assertTrue(n.all(m == n.round(n.array([n.pi/2,0]), 10)))

class TestRadec2eq(unittest.TestCase):
    def testshape(self):
        """Test the ra,dec to equatorial conversion for shape preservation"""
        m = a.coord.radec2eq((0,0))
        self.assertEqual(len(m.shape), 1)
        self.assertEqual(m.shape[0], 3)
        m = a.coord.radec2eq((n.array([0.,0]),n.array([0.,0])))
        self.assertEqual(len(m.shape), 2)
        self.assertEqual(m.shape[0], 3)
        self.assertEqual(m.shape[1], 2)
    def testaccuracy(self):
        """Test the ra,dec to equatorial conversion for accuracy"""
        diag = n.array([[1,0,0],[0,1,0],[0,0,1]], dtype=n.double)
        e1,e2,e3 = diag
        m = n.round(a.coord.radec2eq((0,0)), 10)
        self.assertTrue(n.all(m == e1))
        m = n.round(a.coord.radec2eq((n.pi/2,0)), 10)
        self.assertTrue(n.all(m == e2))
        m = n.round(a.coord.radec2eq((n.pi/2,n.pi/2)), 10)
        self.assertTrue(n.all(m == e3))

class TestLatlong2xyz(unittest.TestCase):
    def testshape(self):
        """Test the lat,long to x,y,z conversion for shape preservation"""
        m = a.coord.latlong2xyz((0,0))
        self.assertEqual(len(m.shape), 1)
        self.assertEqual(m.shape[0], 3)
        m = a.coord.latlong2xyz((n.array([0.,0]),n.array([0.,0])))
        self.assertEqual(len(m.shape), 2)
        self.assertEqual(m.shape[0], 3)
        self.assertEqual(m.shape[1], 2)
    def testaccuracy(self):
        """Test the lat,long to x,y,z conversion for accuracy"""
        diag = n.array([[1,0,0],[0,1,0],[0,0,1]], dtype=n.double)
        e1,e2,e3 = diag
        m = n.round(a.coord.latlong2xyz((0,0)), 10)
        self.assertTrue(n.all(m == e1))
        m = n.round(a.coord.latlong2xyz((n.pi/2,0)), 10)
        self.assertTrue(n.all(m == e3))
        m = n.round(a.coord.latlong2xyz((0,n.pi/2)), 10)
        self.assertTrue(n.all(m == e2))

class TestTop2azalt(unittest.TestCase):
    def testshape(self):
        """Test the x,y,z to az,alt conversion for shape preservation"""
        m = a.coord.top2azalt((0,0,1))
        self.assertEqual(len(m.shape), 1)
        self.assertEqual(m.shape[0], 2)
        m = a.coord.top2azalt((n.array([0.,0,0]),
            n.array([0.,0,0]),n.array([1.,1,1])))
        self.assertEqual(len(m.shape), 2)
        self.assertEqual(m.shape[0], 2)
        self.assertEqual(m.shape[1], 3)
    def testaccuracy(self):
        """Test the x,y,z to az,alt conversion for accuracy"""
        diag = n.array([[1,0,0],[0,1,0],[0,0,1]], dtype=n.double)
        e1,e2,e3 = diag
        m = a.coord.top2azalt(e3)
        self.assertAlmostEqual(m[1], n.pi/2, 10)
        m = n.round(a.coord.top2azalt(e1), 10)
        self.assertTrue(n.all(m == n.round(n.array([n.pi/2,0]), 10)))
        m = n.round(a.coord.top2azalt(e2), 10)
        self.assertTrue(n.all(m == n.round(n.array([0,0]), 10)))

class TestAzalt2top(unittest.TestCase):
    def testshape(self):
        """Test the az,alt to x,y,z conversion for shape preservation"""
        m = a.coord.azalt2top((0,0))
        self.assertEqual(len(m.shape), 1)
        self.assertEqual(m.shape[0], 3)
        m = a.coord.azalt2top((n.array([0.,0]),n.array([0.,0])))
        self.assertEqual(len(m.shape), 2)
        self.assertEqual(m.shape[0], 3)
        self.assertEqual(m.shape[1], 2)
    def testaccuracy(self):
        """Test the az,alt to x,y,z conversion for accuracy"""
        diag = n.array([[1,0,0],[0,1,0],[0,0,1]], dtype=n.double)
        e1,e2,e3 = diag
        m = n.round(a.coord.azalt2top((0,0)), 10)
        self.assertTrue(n.all(m == e2))
        m = n.round(a.coord.azalt2top((n.pi/2,0)), 10)
        self.assertTrue(n.all(m == e1))
        m = n.round(a.coord.azalt2top((0,n.pi/2)), 10)
        self.assertTrue(n.all(m == e3))

class TestEq2top_m(unittest.TestCase):
    def testshape(self):
        """Test the equatorial/x,y,z rotation matrix for shape preservation"""
        m = a.coord.eq2top_m(0, 0)
        self.assertEqual(len(m.shape), 2)
        self.assertEqual(m.shape[0], 3)
        self.assertEqual(m.shape[1], 3)
        m = a.coord.eq2top_m(n.array([0., 0]), n.array([0.,0]))
        self.assertEqual(len(m.shape), 3)
        self.assertEqual(m.shape[0], 2)
        self.assertEqual(m.shape[1], 3)
        self.assertEqual(m.shape[2], 3)
    def testaccuracy(self):
        """Test the equatorial/x,y,z rotation matrix for accuracy"""
        diag = n.array([[1,0,0],[0,1,0],[0,0,1]], dtype=n.double)
        e1,e2,e3 = diag
        m = n.round(a.coord.eq2top_m(0, 0), 10)
        self.assertTrue(n.all(m == n.array([e2,e3,e1])))
        m = n.round(a.coord.eq2top_m(-n.pi/2, 0.), 10)
        self.assertTrue(n.all(m == n.array([-e1,e3,e2])))
        m = n.round(a.coord.eq2top_m(0, n.pi/2), 10)
        self.assertTrue(n.all(m == n.array([e2,-e1,e3])))
        
class TestTop2eq_m(unittest.TestCase):
    def testshape(self):
        """Test the x,y,z/equatorial rotation matrix for shape preservation"""
        m = a.coord.top2eq_m(0, 0)
        self.assertEqual(len(m.shape), 2)
        self.assertEqual(m.shape[0], 3)
        self.assertEqual(m.shape[1], 3)
        m = a.coord.top2eq_m(n.array([0., 0]), n.array([0.,0]))
        self.assertEqual(len(m.shape), 3)
        self.assertEqual(m.shape[0], 2)
        self.assertEqual(m.shape[1], 3)
        self.assertEqual(m.shape[2], 3)
    def testaccuracy(self):
        """Test the x,y,z/equatorial rotation matrix for accuracy"""
        diag = n.array([[1,0,0],[0,1,0],[0,0,1]], dtype=n.double)
        e1,e2,e3 = diag
        m = n.round(a.coord.top2eq_m(0, 0), 10)
        self.assertTrue(n.all(m == n.array([e3,e1,e2])))
        m = n.round(a.coord.top2eq_m(-n.pi/2, 0.), 10)
        self.assertTrue(n.all(m == n.array([-e1,e3,e2])))
        m = n.round(a.coord.top2eq_m(0, n.pi/2), 10)
        self.assertTrue(n.all(m == n.array([-e2,e1,e3])))

class TestSuite(unittest.TestSuite):
    """A unittest.TestSuite class which contains all of the aipy.coord unit tests."""

    def __init__(self):
        unittest.TestSuite.__init__(self)

        loader = unittest.TestLoader()
        self.addTests(loader.loadTestsFromTestCase(TestConvert))
        self.addTests(loader.loadTestsFromTestCase(TestConvert_m))
        self.addTests(loader.loadTestsFromTestCase(TestRot_m))
        self.addTests(loader.loadTestsFromTestCase(TestXyz2thphi))
        self.addTests(loader.loadTestsFromTestCase(TestThphi2xyz))
        self.addTests(loader.loadTestsFromTestCase(TestEq2radec))
        self.addTests(loader.loadTestsFromTestCase(TestRadec2eq))
        self.addTests(loader.loadTestsFromTestCase(TestLatlong2xyz))
        self.addTests(loader.loadTestsFromTestCase(TestTop2azalt))
        self.addTests(loader.loadTestsFromTestCase(TestAzalt2top))
        self.addTests(loader.loadTestsFromTestCase(TestEq2top_m))
        self.addTests(loader.loadTestsFromTestCase(TestTop2eq_m))

if __name__ == '__main__':
    unittest.main()



#!/usr/bin/python
"""
Detect and flag RFI related effects in UV files.  Uses statistical thresholding
to identify outliers in power and narrowness of features.  Does an improved
job of identifying low-level interference if sky model and crosstalk are
removed from the data first.
"""

import numpy as np, aipy as a, os, sys, pickle, optparse

o = optparse.OptionParser()
o.set_usage('xrfi.py [options] *.uv')
o.set_description(__doc__)
o.add_option('-c', '--chan', dest='chan', 
    help='Manually flag channels before xrfi processing.  Options are "<chan1 #>,..." (a list of channels), or "<chan1 #>_<chan2 #>" (a range of channels).  Default is None.')
o.add_option('-n', '--nsig', dest='nsig', default=2., type='float',
    help='Number of standard deviations above mean to flag.  Default 2.')
o.add_option('-m', '--flagmode', dest='flagmode', default='both',
    help='Can be val,int,both,none for flagging by value only, integration only, both, or only manually flagged channels.  Default both.')
o.add_option('-s', '--share', dest='share', action='store_true',
    help='Flag a channel, integration if any pol/baseline flags it (share flags).')
o.add_option('--ch_thresh', dest='ch_thresh',type='float',default=.33,
    help='Fraction of the data in a channel which, if flagged, will result in the entire channel being flagged.  Default .33')
o.add_option('--int_thresh', dest='int_thresh',type='float',default=.99,
    help='Fraction of the data in an integration which, if flagged, will result in the entire integration being flagged.  Default .99')
o.add_option('--raw', dest='raw', action='store_true',
    help='Flag by integration without removing a smooth function.')
o.add_option('--reflag', dest='reflag', action='store_true',
    help='Ignore any previous flagging.')
opts, args = o.parse_args(sys.argv[1:])

def share_mask(mask):
    '''Flag a chan,integration if any pol/baseline has flagged it.'''
    m = {}
    for pol in mask:
      for bl in mask[pol]:
        for t in mask[pol][bl]:
            m[t] = m.get(t,0) | mask[pol][bl]
    for pol in mask:
      for bl in mask[pol]:
        for t in mask[pol][bl]:
            mask[pol][bl][t] = m[t]
    return mask

# Parse command-line options
uv = a.miriad.UV(args[0])
if not opts.chan is None:
    chans = a.scripting.parse_chans(opts.chan, uv['nchan'])
else:
    chans = []
    opts.chan = 'None'
del(uv)

for uvfile in args:
    uvofile = uvfile+'r'
    print uvfile,'->',uvofile
    if os.path.exists(uvofile):
        print uvofile, 'exists, skipping.'
        continue
    uvi = a.miriad.UV(uvfile)
    (uvw,jd,(i,j)),d,f = uvi.read(raw=True)
    uvi.rewind()
    # Gather all data and each time step
    window = None
    data,mask,times = {}, {}, []
    for (uvw,t,(i,j)), d, f in uvi.all(raw=True):
        if len(times) == 0 or times[-1] != t: times.append(t)
        pol = uvi['pol']
        if not pol in data:
            data[pol] = {}
            mask[pol] = {}
        bl = a.miriad.ij2bl(i,j)
        if not bl in data[pol]:
            data[pol][bl] = {}
            mask[pol][bl] = {}
        if opts.reflag: f = np.zeros_like(f)
        f = np.where(np.abs(d) == 0, 1, f)
        if i != j:
            if window is None: 
                window = d.size/2 - abs(np.arange(d.size) - d.size/2)
            d = np.fft.fft(np.fft.ifft(d) * window)
        # Manually flagged channels
        f[chans] = 1
        mask[pol][bl][t] = f
        data[pol][bl][t] = d

    # Generate a single mask for all baselines which masks data if any
    # baseline has an outlier at that freq/time.  Data flagged
    # strictly on basis of nsigma above mean.
    if opts.flagmode in ['val', 'both']:
        for pol in data:
          for bl in data[pol]:
            i, j = a.miriad.bl2ij(bl)
            if i == j: continue
            data_times = data[pol][bl].keys()
            d = np.ma.array([data[pol][bl][t] for t in data_times],
                mask=[mask[pol][bl][t] for t in data_times])
            hi_thr, lo_thr = a.rfi.gen_rfi_thresh(d, nsig=opts.nsig)
            m = np.where(np.abs(d) > hi_thr,1,0)
            for i, t in enumerate(data_times): mask[pol][bl][t] |= m[i]
            # If more than ch_thresh of the data in a channel or 
            # integration is flagged, flag the whole thing
            ch_cnt = np.array([mask[pol][bl][t] for t in data_times]).sum(axis=0)
            ch_msk = np.where(ch_cnt > ch_cnt.max()*opts.ch_thresh,1,0)
            for t in mask[pol][bl]:
                if mask[pol][bl][t].sum() > mask[pol][bl][t].size \
                        * opts.int_thresh:
                    mask[pol][bl][t] |= 1
                else: mask[pol][bl][t] |= ch_msk

    # Use autocorrelations to flag entire integrations which have
    # anomalous powers.  At least 2 antennas must agree for a 
    # integration to get flagged.
    if opts.flagmode in ['int', 'both']:
        for pol in data:
          for bl in data[pol]:
            i, j = a.miriad.bl2ij(bl)
            if i != j: continue
            data_times = data[pol][bl].keys()
            data_times.sort()
            d = np.ma.array([data[pol][bl][t] for t in data_times],
                mask=[mask[pol][bl][t] for t in data_times])
            bad_ints = a.rfi.flag_by_int(d, nsig=opts.nsig, raw=opts.raw)
            for i in np.where(bad_ints)[0]:
                t = data_times[i]
                mask[pol][bl][t] |= 1

    if opts.share: mask = share_mask(mask)
    # Generate a pipe for applying the mask to data as it comes in.
    def rfi_mfunc(uv, preamble, data, flags):
        uvw, t, (i,j) = preamble
        f = mask[uv['pol']][a.miriad.ij2bl(i,j)][t]
        return preamble, np.where(f, 0, data), f

    uvi.rewind()
    uvo = a.miriad.UV(uvofile, status='new')
    uvo.init_from_uv(uvi)
    uvo.pipe(uvi, mfunc=rfi_mfunc, raw=True, append2hist='XRFI: nsig=%f chans=%s mode=%s ch_thresh=%f int_thresh=%f raw=%s reflag=%s share=%s\n' %  (opts.nsig, opts.chan, opts.flagmode, opts.ch_thresh, opts.int_thresh, opts.raw, opts.reflag, opts.share))



'''This module provides a front-end interface for accessing sources in 
all catalogs in the _src module of AIPY.'''
import fit, _src

def get_catalog(srcs=None, cutoff=None, catalogs=['helm','misc']):
    """Return a source catalog created out of the sources listed in 'srcs',
    or with fluxes above the specified (jy_cutoff,freq_ghz) in 'cutoff'.
    Searches catalogs listed in 'catalogs'."""
    srclist = []
    for c in catalogs:
        try: c = getattr(_src, c)
        except(AttributeError): continue
        srclist += c.get_srcs(srcs=srcs, cutoff=cutoff)
    # Add in sources that are already made
    if srcs != None: srclist += [s for s in srcs if type(s) != str]
    return fit.SrcCatalog(srclist)



# -*- coding: utf-8 -*-
import tempfile
import unittest, numpy as np, os
import aipy.miriad as m, aipy._miriad as _m

class TestMiriadUV(unittest.TestCase):
    def setUp(self):
        self.tmppath = tempfile.mkdtemp(prefix='miriad-test-', suffix='.tmp')
        self.filename1 = os.path.join(self.tmppath, 'test1.uv')
        self.filename2 = os.path.join(self.tmppath, 'test2.uv')
        uv = m.UV(self.filename1, status='new', corrmode='r')
        uv['history'] = 'Made this file from scratch.\n'
        uv.add_var('nchan', 'i')
        uv.add_var('pol', 'i')
        uv['nchan'] = 4
        uv['pol'] = -5
        uvw = np.array([1,2,3], dtype=np.double)
        preamble = (uvw, 12345.6789, (0,1))
        self.data = np.ma.array([1j,2,3j,4], mask=[0,0,1,0], dtype=np.complex64)
        uv.write(preamble,self.data)
        uv['pol'] = -6
        uv.write(preamble,self.data)
    def test_maxchan(self):
        self.assertEqual(type(_m.MAXCHAN), int)
    def test_immediate_corr(self):
        """Test immediate corr of a Miriad UV file"""
        uv = m.UV(self.filename2, status='new')
        self.assertEqual(uv.vartable['corr'], 'r')
    def test_vartable(self):
        """Test accesing vartable data in a Miriad UV file"""
        uv = m.UV(self.filename1)
        self.assertEqual(uv.vartable['corr'], 'r')
        self.assertEqual(uv.vartable['nchan'], 'i')
        self.assertEqual(uv.vartable['pol'], 'i')
    def test_data(self):
        """Test writing data from a Miriad UV file"""
        uv = m.UV(self.filename1)
        self.assertEqual(uv['history'], 'Made this file from scratch.\n')
        (uvw,t,bl),d = uv.read()
        self.assertEqual(uv['nchan'], 4)
        self.assertEqual(uv['pol'], -5)
        self.assertEqual(bl, (0,1))
        self.assertEqual(t, 12345.6789)
        self.assertTrue(np.all(uvw == np.array([1,2,3], dtype=np.double)))
        self.assertTrue(np.all(d == self.data))
        (uvw,t,bl),d = uv.read()
        self.assertEqual(uv['nchan'], 4)
        self.assertEqual(uv['pol'], -6)
        self.assertEqual(bl, (0,1))
        self.assertEqual(t, 12345.6789)
        self.assertTrue(np.all(uvw == np.array([1,2,3], dtype=np.double)))
        self.assertTrue(np.all(d == self.data))
    def tearDown(self):
        os.system("rm -rf %s" % self.tmppath)

class TestMiriadUV_corrj(unittest.TestCase):
    def setUp(self):
        self.tmppath = tempfile.mkdtemp(prefix='miriad-test-', suffix='.tmp')
        self.filename1 = os.path.join(self.tmppath, 'test1.uv')
        uv = m.UV(self.filename1, status='new', corrmode='j')
        uv['history'] = 'Made this file from scratch.\n'
        uv.add_var('nchan', 'i')
        uv.add_var('pol', 'i')
        uv['nchan'] = 4
        uv['pol'] = -5
        uvw = np.array([1,2,3], dtype=np.double)
        preamble = (uvw, 12345.6789, (0,1))
        self.data = np.ma.array([1j,2,3j,4], mask=[0,0,1,0], dtype=np.complex64)
        uv.write(preamble,self.data)
        uv['pol'] = -6
        uv.write(preamble,self.data)
    def test_vartable(self):
        """Test accesing vartable data in a Miriad UV file"""
        uv = m.UV(self.filename1, corrmode='j')
        self.assertEqual(uv.vartable['corr'], 'j')
        self.assertEqual(uv.vartable['nchan'], 'i')
        self.assertEqual(uv.vartable['pol'], 'i')
    def test_data(self):
        """Test writing data from a Miriad UV file"""
        uv = m.UV(self.filename1)
        self.assertEqual(uv['history'], 'Made this file from scratch.\n')
        (uvw,t,bl),d = uv.read()
        self.assertEqual(uv['nchan'], 4)
        self.assertEqual(uv['pol'], -5)
        self.assertEqual(bl, (0,1))
        self.assertEqual(t, 12345.6789)
        self.assertTrue(np.all(uvw == np.array([1,2,3], dtype=np.double)))
        self.assertTrue(np.all(np.abs(d - self.data) < 1e-4))
        (uvw,t,bl),d = uv.read()
        self.assertEqual(uv['nchan'], 4)
        self.assertEqual(uv['pol'], -6)
        self.assertEqual(bl, (0,1))
        self.assertEqual(t, 12345.6789)
        self.assertTrue(np.all(uvw == np.array([1,2,3], dtype=np.double)))
        self.assertTrue(np.all(np.abs(d - self.data) < 1e-4))
    def tearDown(self):
        os.system("rm -rf %s" % self.tmppath)

class TestSuite(unittest.TestSuite):
    """A unittest.TestSuite class which contains all of the aipy.miriad unit tests."""

    def __init__(self):
        unittest.TestSuite.__init__(self)

        loader = unittest.TestLoader()
        self.addTests(loader.loadTestsFromTestCase(TestMiriadUV))

if __name__ == '__main__':
    unittest.main()



'''The GB6 Catalog.
Data files are in tab-separated format from Vizier.
To download in the correct format, open a catalog online in Vizier,
select'Tab-Separated-Values' as the Output layout in the drop-down box, set
the maximum entries to 'unlimited', and click 'Sexagesimal' under the box
for 'Target Name or Position'.  Submit the query, and copy the output to a
txt file.  Copy this file to "gb6.txt" in the _src directory of your AIPY
installation.'''

import aipy as a, numpy as np, os

class GBSixCatalog(a.fit.SrcCatalog):
    def fromfile(self,filename):
        f = open(filename)
        addsrcs = []
        for L in [L for L in f.readlines() if not L.startswith('#')]:
            text = L.split('\t')
            if len(text) <= 4: continue
            try: int(text[0][0])
            except(ValueError): continue
            ra = text[0].replace(' ',':')
            dec = text[1].replace(' ',':')
            name = text[18].strip()
            jys = float(text[8])/1000.
            addsrcs.append(a.fit.RadioFixedBody(ra, dec, name=name,
                jys=jys, index=0, mfreq=4.85))
        self.add_srcs(addsrcs)
            
GBSIXFILE = os.path.dirname(__file__) + os.sep + 'gb6.txt'
_gbsixcat = None

def get_srcs(srcs=None, cutoff=None):
	global _gbsixcat
	if _gbsixcat is None:
		_gbsixcat = GBSixCatalog()
        _gbsixcat.fromfile(GBSIXFILE)
	if srcs is None:
		if cutoff is None: srcs = _gbsixcat.keys()
		else:
			cut, fq = cutoff
			fq = np.array([fq])
			for s in _gbsixcat.keys(): _gbsixcat[s].update_jys(fq)
			srcs = [s for s in _gbsixcat.keys() if _gbsixcat[s].jys[0] > cut]
	srclist = []
	for s in srcs:
		try: srclist.append(_gbsixcat[s])
		except(KeyError): pass
	return srclist



import unittest, ephem, random
import aipy as a, numpy as n

DIM = 128

class TestClean(unittest.TestCase):
    def test_clean2dc(self):
        res = n.zeros((DIM,DIM), dtype=n.complex)
        ker = n.zeros((DIM,DIM), dtype=n.complex)
        mdl = n.zeros((DIM,DIM), dtype=n.complex)
        area = n.zeros((DIM,DIM), dtype=n.int)
        self.assertRaises(ValueError, a._deconv.clean, \
            res,ker,mdl,area.astype(n.float))
        ker[0,0] = 1.
        res[0,0] = 1.; res[5,5] = 1.
        area[:4][:4] = 1
        rv = a._deconv.clean(res,ker,mdl,area,tol=1e-8)
        self.assertAlmostEqual(res[0,0], 0, 3)
        self.assertEqual(res[5,5], 1)
    def test_clean2dr(self):
        res = n.zeros((DIM,DIM), dtype=n.float)
        ker = n.zeros((DIM,DIM), dtype=n.float)
        mdl = n.zeros((DIM,DIM), dtype=n.float)
        area = n.zeros((DIM,DIM), dtype=n.int)
        self.assertRaises(ValueError, a._deconv.clean, \
            res,ker,mdl,area.astype(n.float))
        ker[0,0] = 1.
        res[0,0] = 1.; res[5,5] = 1.
        area[:4][:4] = 1
        rv = a._deconv.clean(res,ker,mdl,area,tol=1e-8)
        self.assertAlmostEqual(res[0,0], 0, 3)
    def test_clean1dc(self):
        res = n.zeros((DIM,), dtype=n.complex)
        ker = n.zeros((DIM,), dtype=n.complex)
        mdl = n.zeros((DIM,), dtype=n.complex)
        area = n.zeros((DIM,), dtype=n.int)
        self.assertRaises(ValueError, a._deconv.clean, \
            res,ker,mdl,area.astype(n.float))
        ker[0] = 1.
        res[0] = 1.; res[5] = 1.
        area[:4] = 1
        rv = a._deconv.clean(res,ker,mdl,area,tol=1e-8)
        self.assertAlmostEqual(res[0], 0, 3)
        self.assertEqual(res[5], 1)
    def test_clean1dr(self):
        res = n.zeros((DIM,), dtype=n.float)
        ker = n.zeros((DIM,), dtype=n.float)
        mdl = n.zeros((DIM,), dtype=n.float)
        area = n.zeros((DIM,), dtype=n.int)
        self.assertRaises(ValueError, a._deconv.clean, \
            res,ker,mdl,area.astype(n.float))
        ker[0] = 1.
        res[0] = 1.; res[5] = 1.
        area[:4] = 1
        rv = a._deconv.clean(res,ker,mdl,area,tol=1e-8)
        self.assertAlmostEqual(res[0], 0, 3)
        self.assertEqual(res[5], 1)
    def test_clean2d_stop_if_div(self):
        DIM1,DIM2 = 1000, 250
        dim = n.random.normal(size=(DIM1,DIM2))
        dbm = n.random.normal(size=(DIM1,DIM2))
        mdl = n.zeros(dim.shape, dtype=dim.dtype)
        area = n.ones(dim.shape, dtype=n.int)
        rv = a._deconv.clean(dim, dbm, mdl, area, gain=.1, tol=1e-2, stop_if_div=0, maxiter=100)
    

if __name__ == '__main__':
    unittest.main()



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
import aipy as a, numpy as n
#import pylab as p

class TestDeconv(unittest.TestCase):
    def setUp(self):
        SIZE = 100
        NOISE = .001
        i = n.zeros((SIZE,SIZE), n.float)
        i[10,10] = 10.
        i[20:25,20:25] = 1.
        i[30:40,30:40] = .1
        self.b = a.img.gaussian_beam(2, shape=i.shape)
        self.b[0,0] = .05

        self.d = n.abs(n.fft.ifft2(n.fft.fft2(i) * n.fft.fft2(self.b)))
        ns = n.random.normal(scale=NOISE, size=i.shape)
        self.d = n.abs(self.d + ns)
        
    def test_clean(self):
        """Test that the standard clean deconvolution runs"""
        #print 'Clean'
        #p.subplot(221)
        c,info = a.deconv.clean(self.d, self.b, verbose=False)
        #p.title('CLEAN')
        #p.imshow(n.log10(c), vmin=-5, vmax=1)

    def test_lsq(self):
        """Test that least squared deconvolution runs"""
        #print 'LSQ'
        #p.subplot(222)
        c,info = a.deconv.lsq(self.d, self.b, verbose=False)
        #p.title('LSQ')
        #p.imshow(n.log10(c), vmin=-5, vmax=1)
        
    def test_mem(self):
        """Test the maximum entropy deconvolution runs"""
        #print 'MEM'
        #p.subplot(223)
        c,info = a.deconv.maxent(self.d, self.b, n.var(self.d**2)*.5, verbose=False)
        #p.title('MEM')
        #p.imshow(n.log10(c), vmin=-5, vmax=1)

    def test_anneal(self):
        """Test that simulated annealing deconvolution runs"""
        #print 'Anneal'
        #p.subplot(224)
        c,info = a.deconv.anneal(self.d, self.b, verbose=False)
        #p.title('Anneal')
        #p.imshow(n.log10(c), vmin=-5, vmax=1)
        #p.colorbar()
        #p.show()

class TestSuite(unittest.TestSuite):
    """A unittest.TestSuite class which contains all of the aipy.deconv unit tests."""

    def __init__(self):
        unittest.TestSuite.__init__(self)

        loader = unittest.TestLoader()
        self.addTests(loader.loadTestsFromTestCase(TestDeconv))

if __name__ == '__main__':
    unittest.main()



"""
A pure-python, spline-like implementation of 1D interpolation.  Uses
an FIR filter (spline) to interpolate, and a polynomial to extrapolate
boundaries.

Author: Aaron Parsons
Date: 12/12/07
Revisions:
"""

__all__ = ['interpolate', 'default_filter']

import numpy as np

def subsample(a, factor):
    rv = np.zeros((a.size, factor), dtype=a.dtype)
    rv[:,0] = a
    wgts = np.zeros((a.size, factor), dtype=np.float)
    wgts[:,0] = 1.
    return rv.flatten(), wgts.flatten()

def polyextend(y, nsamples, degree=2):
    """Extend y on either end by nsamples with a polynomial of the specified
    degree.  Assumes y is sampled on a uniform x grid."""
    x = np.arange(-nsamples,nsamples)
    p = np.polyfit(x[-nsamples:],y[:nsamples],degree)
    y0 = np.polyval(p, x[:nsamples])
    x = np.arange(-nsamples+1,nsamples+1)
    p = np.polyfit(x[:nsamples],y[-nsamples:],degree)
    y1 = np.polyval(p, x[-nsamples:])
    return np.concatenate([y0, y, y1])
    
def default_filter(xs, freq=.25):
    """A basic smoothing filter using a Hamming window and a sinc function
    of the specified frequency.  Input a sample grid [-x,x] to get the
    FIR coefficients used for smoothing."""
    i = np.arange(xs.size, dtype=np.float) / xs.size * 2*np.pi
    return (1-np.cos(i)) * np.sinc(freq*xs)

def interpolate(ys, factor, filter=default_filter, order=4):
    """Oversample ys by the specified factor using a filter function to
    interpolated between samples.  The filter function will be used to
    construct an FIR filter for x values [-order,order] in steps of 1/order.
    Order should be even to make this an averaging filter.  Internally,
    ys is extended by 2nd degree polynomial in both directions to attempt
    a smooth boundary transition."""
    step = 1./factor
    ys = polyextend(ys, order)
    ys_ss, wgts = subsample(ys, factor)
    fir = filter(np.arange(-order,order+step,step))
    ys_ss = np.convolve(ys_ss, fir, mode='valid')
    wgts = np.convolve(wgts, fir, mode='valid')
    return ys_ss / wgts
   



#! /usr/bin/env python
"""
Add or subtract the data in the second UV file from the data in the 
first UV file.  They'll need to have corresponding integrations/data order.
"""

import aipy as a, numpy as np, sys, optparse, os

o = optparse.OptionParser()
o.set_usage('uv_addsub.py [options] file1.uv file2.uv')
o.set_description(__doc__)
o.add_option('--sub', dest='sub', action='store_true',
    help='Subtract the data instead of adding.')
opts,args = o.parse_args(sys.argv[1:])

assert(len(args) == 2)
uv1 = a.miriad.UV(args[0])
uv2 = a.miriad.UV(args[1])

def mfunc(uv, p, d, f):
    p2,d2,f2 = uv2.read(raw=True)
    f = np.logical_or(f,f2)
    if opts.sub: d = d - d2
    else: d = d + d2
    return p, np.where(f, 0, d), f

if opts.sub: filename = args[0] + 'd'
else: filename = args[0] + 'a'
if os.path.exists(filename):
    print 'File exists: skipping'
    sys.exit(0)
print args[0], '->', filename
uvo = a.miriad.UV(filename, status='new')
uvo.init_from_uv(uv1)
uvo.pipe(uv1, mfunc=mfunc, raw=True, 
    append2hist='UVADDSUB: --sub=%s file=%s\n' % (opts.sub, args[1]))



#! /usr/bin/env python
"""
Helps you figure out which version of AIPY you have installed.
Lists:
    version
    install location
    last commit
    branch
"""
import aipy
print "AIPY base version in use:",
try:
    print aipy.__version__
except(AttributeError):
    print "Version number not Found"
print "Install location:",aipy.__file__
print "Last Git commit log:"
try:
    print aipy.__gitlog__
except(AttributeError):
    print "     Gitlog not found"
print "Branch:",
try:
    print aipy.__branch__
except(AttributeError):
    print "Branch Unknown"




"""
Module for reading and setting parameters in components of an AntennaArray
simulation for purpose of fitting.
"""
import amp, numpy as np

#  _   _ _   _ _ _ _           _____                 _   _                 
# | | | | |_(_) (_) |_ _   _  |  ___|   _ _ __   ___| |_(_) ___  _ __  ___ 
# | | | | __| | | | __| | | | | |_ | | | | '_ \ / __| __| |/ _ \| '_ \/ __|
# | |_| | |_| | | | |_| |_| | |  _|| |_| | | | | (__| |_| | (_) | | | \__ \
#  \___/ \__|_|_|_|\__|\__, | |_|   \__,_|_| |_|\___|\__|_|\___/|_| |_|___/
#                      |___/        

def flatten_prms(prms, prm_list=None):
    """Generate list of parameters suitable for passing to fitting
    algorithm from heirarchical parameter dictionary 'prms', along
    with 'key_list' information for reconstructing such a dictionary from
    a list.  'prm_list' is only for recursion."""
    if prm_list is None: prm_list = []
    key_list = {}
    keys = prms.keys()
    keys.sort()
    for k in keys:
        if type(prms[k]) == dict:
            prm_list, new_key_list = flatten_prms(prms[k], prm_list)
            key_list[k] = new_key_list
        else:
            try:
                key_list[k] = (len(prm_list), len(prms[k]))
                prm_list += list(prms[k])
            except(TypeError):
                key_list[k] = (len(prm_list), 1)
                prm_list.append(prms[k])
    return prm_list, key_list

def reconstruct_prms(prm_list, key_list):
    """Generate a heirarchical parameter dictionary from parameter
    list (prm_list) and 'key_list' information from flatten_prms."""
    prms = {}
    for k in key_list:
        v = key_list[k]
        if type(v) == dict: prms[k] = reconstruct_prms(prm_list, v)
        else:
            i, L = v
            if L > 1: prms[k] = prm_list[i:i+L]
            else: prms[k] = prm_list[i]
    return prms

def print_params(prms, indent='', grad=None):
    """Print nice looking representation of a parameter dictionary."""
    keys = prms.keys()
    keys.sort()
    for k in keys:
        v = prms[k]
        if (type(v) is dict and v == {}) or v is None or \
                (type(v) is list and v == []):
            continue
        if type(v) == dict:
            print indent, k
            if grad is None: print_params(v, indent + '  ')
            else: print_params(v, indent + '  ', grad[k])
        else:
            print indent, k,
            if grad is None:
                if not type(v) is list:
                    try: v = [list(v)]
                    except(TypeError): v = [v]
                if len(v) == 1: print v[0]
                else:
                    print
                    for i in v: print indent, ' ', i
            else:
                print
                print indent, v, '\t<', grad[k], '>'
                if not type(v) is list:
                    try: v = [list(v)]
                    except(TypeError): v = [v]
                for i in len(v):
                    print indent, ' ', v[i], '\t<', grad[k][i], '>'


#  ____           _ _       _____ _              _ ____            _       
# |  _ \ __ _  __| (_) ___ |  ___(_)_  _____  __| | __ )  ___   __| |_   _ 
# | |_) / _` |/ _` | |/ _ \| |_  | \ \/ / _ \/ _` |  _ \ / _ \ / _` | | | |
# |  _ < (_| | (_| | | (_) |  _| | |>  <  __/ (_| | |_) | (_) | (_| | |_| |
# |_| \_\__,_|\__,_|_|\___/|_|   |_/_/\_\___|\__,_|____/ \___/ \__,_|\__, |
#                                                                    |___/ 

class RadioFixedBody(amp.RadioFixedBody):
    """Class representing a source at fixed RA,DEC.  Adds get_params() and
    set_params() to amp.RadioFixedBody."""
    def get_params(self, prm_list=['*']):
        """Return all fitable parameters in a dictionary."""
        aprms = {
            'jys':      float(self._jys),
            'index':    float(self.index),
            'ra':       float(self._ra),
            'dec':      float(self._dec),
            'a1':       float(self.srcshape[0]),
            'a2':       float(self.srcshape[1]),
            'th':       float(self.srcshape[2]),
            'dra':      float(self.ionref[0]),
            'ddec':     float(self.ionref[1]),
            'mfreq':    float(self.mfreq),
        }
        prms = {}
        for p in prm_list:
            if p.startswith('*'): return aprms
            try: prms[p] = aprms[p]
            except(KeyError): pass
        return prms
    def set_params(self, prms):
        """Set all parameters from a dictionary."""
        try: self._jys = prms['jys']
        except(KeyError): pass
        try: self.index = prms['index']
        except(KeyError): pass
        try: self._ra = prms['ra']
        except(KeyError): pass
        try: self._dec = prms['dec']
        except(KeyError): pass
        try: self.srcshape[0] = prms['a1']
        except(KeyError): pass
        try: self.srcshape[1] = prms['a2']
        except(KeyError): pass
        try: self.srcshape[2] = prms['th']
        except(KeyError): pass
        try: self.ionref[0] = prms['dra']
        except(KeyError): pass
        try: self.ionref[1] = prms['ddec']
        except(KeyError): pass

#  ____           _ _      ____                  _       _ 
# |  _ \ __ _  __| (_) ___/ ___| _ __   ___  ___(_) __ _| |
# | |_) / _` |/ _` | |/ _ \___ \| '_ \ / _ \/ __| |/ _` | |
# |  _ < (_| | (_| | | (_) |__) | |_) |  __/ (__| | (_| | |
# |_| \_\__,_|\__,_|_|\___/____/| .__/ \___|\___|_|\__,_|_|
#                               |_|                        

class RadioSpecial(amp.RadioSpecial):
    """Class representing moving sources (Sun,Moon,planets). Adds get_params() 
    and set_params() to amp.RadioSpecial."""
    def get_params(self, prm_list=['*']):
        """Return all fitable parameters in a dictionary."""
        aprms = {
            'jys':      float(self._jys),
            'index':    float(self.index),
            'a1':       float(self.srcshape[0]),
            'a2':       float(self.srcshape[1]),
            'th':       float(self.srcshape[2]),
            'dra':      float(self.ionref[0]),
            'ddec':     float(self.ionref[1]),
        }
        prms = {}
        for p in prm_list:
            if p.startswith('*'): return aprms
            try: prms[p] = aprms[p]
            except(KeyError): pass
        return prms
    def set_params(self, prms):
        """Set all parameters from a dictionary."""
        try: self._jys = prms['jys']
        except(KeyError): pass
        try: self.index = prms['index']
        except(KeyError): pass
        try: self.srcshape[0] = prms['a1']
        except(KeyError): pass
        try: self.srcshape[1] = prms['a2']
        except(KeyError): pass
        try: self.srcshape[2] = prms['th']
        except(KeyError): pass
        try: self.ionref[0] = prms['dra']
        except(KeyError): pass
        try: self.ionref[1] = prms['ddec']
        except(KeyError): pass

#  ____            ____      _        _             
# / ___| _ __ ___ / ___|__ _| |_ __ _| | ___   __ _ 
# \___ \| '__/ __| |   / _` | __/ _` | |/ _ \ / _` |
#  ___) | | | (__| |__| (_| | || (_| | | (_) | (_| |
# |____/|_|  \___|\____\__,_|\__\__,_|_|\___/ \__, |
#                                             |___/ 

class SrcCatalog(amp.SrcCatalog):
    """Class for holding a catalog of celestial sources.  Adds get_params()
    and set_params() to amp.SrcCatalog."""
    def get_params(self, src_prms={'*':'*'}):
        """Return all fitable parameters in a dictionary."""
        prms = {}
        for k in src_prms:
            if k.startswith('*'): srcs = self.keys()
            else: srcs = [k]
            prm_list = src_prms[k]
            if type(prm_list) is str: prm_list = [prm_list]
            for s in srcs:
                try: prms[s] = self[s].get_params(prm_list)
                except(KeyError): pass
        return prms
    def set_params(self, prms):
        """Set all parameters from a dictionary."""
        for s in prms:
            try: self[s].set_params(prms[s])
            except(KeyError): pass

#  ____
# | __ )  ___  __ _ _ __ ___
# |  _ \ / _ \/ _` | '_ ` _ \
# | |_) |  __/ (_| | | | | | |
# |____/ \___|\__,_|_| |_| |_|

class Beam(amp.Beam):
    """Representation of a flat (gain=1) antenna beam pattern."""
    def get_params(self, prm_list=['*']):
        return {}
    def set_params(self, prms):
        return False

class Beam2DGaussian(amp.Beam2DGaussian):
    """Representation of a 2D Gaussian beam pattern, with default setting for 
    a flat beam."""
    def get_params(self, prm_list=['*']):
        """Return all fitable parameters in a dictionary."""
        aprms = {'bm_xwidth':self.xwidth, 'bm_ywidth':self.ywidth}
        prms = {}
        for p in prm_list:
            if p.startswith('*'): return aprms
            try: prms[p] = aprms[p]
            except(KeyError): pass
        return prms
    def set_params(self, prms):
        """Set all parameters from a dictionary."""
        changed = False
        try: self.xwidth, changed = prms['bm_xwidth'], True
        except(KeyError): pass
        try: self.ywidth, changed = prms['bm_ywidth'], True
        except(KeyError): pass
        if changed: self.update()
        return changed

class BeamPolynomial(amp.BeamPolynomial):
    """Representation of a gaussian beam model whose width varies with azimuth
    angle and with frequency."""
    def get_params(self, prm_list=['*']):
        """Return all fitable parameters in a dictionary."""
        aprms = {'bm_poly':self.poly.flatten()}
        prms = {}
        for p in prm_list:
            if p.startswith('*'): return aprms
            try: prms[p] = aprms[p]
            except(KeyError): pass
        return prms
    def set_params(self, prms):
        """Set all parameters from a dictionary."""
        changed = False
        try: 
            poly_azfreq = prms['bm_poly']
            poly_azfreq.shape = self.poly.shape
            self.poly = poly_azfreq
            changed = True
        except(KeyError): pass
        if changed: self.update()
        return changed

class BeamAlm(amp.BeamAlm):
    """Representation of a beam model where each pointing has a response
    defined as a polynomial in frequency, and the spatial distributions of 
    these coefficients decomposed into spherical harmonics."""
    def get_params(self, prm_list=['*']):
        """Return all fitable parameters in a dictionary."""
        aprms = {}
        for i, a in enumerate(self.alm):
            k = 'alm%d' % i
            data = a.get_data()
            aprms[k] = np.array([data.real, data.imag]).transpose().flatten()
        prms = {}
        for p in prm_list:
            if p.startswith('*'): prms = aprms
            try: prms[p] = aprms[p]
            except(KeyError): pass
        return prms
    def set_params(self, prms):
        """Set all parameters from a dictionary."""
        changed = False
        for p in prms:
            if not p.startswith('alm'): continue
            changed = True
            data = np.array(prms[p])
            c = int(p[3:])
            data.shape = (data.size/2, 2)
            data = data[:,0] + data[:,1] * 1j
            if c < len(self.alm): self.alm[-1-c].set_data(data)
        if changed: self.update()
        return changed

#     _          _                         
#    / \   _ __ | |_ ___ _ __  _ __   __ _ 
#   / _ \ | '_ \| __/ _ \ '_ \| '_ \ / _` |
#  / ___ \| | | | ||  __/ | | | | | | (_| |
# /_/   \_\_| |_|\__\___|_| |_|_| |_|\__,_|

class Antenna(amp.Antenna):
    """Representation of physical location and beam pattern of individual 
    antenna in array.  Adds get_params() and set_params() to amp.Antenna."""
    def get_params(self, prm_list=['*']):
        """Return all fitable parameters in a dictionary."""
        x,y,z = self.pos
        aprms = {'x':x, 'y':y, 'z':z, 'dly':self._phsoff[-2], 
            'off':self._phsoff[-1], 'phsoff':self._phsoff}
        aprms['bp_r'] = list(self.bp_r)
        aprms['bp_i'] = list(self.bp_i)
        aprms['amp'] = self.amp
        aprms.update(self.beam.get_params(prm_list))
        prms = {}
        for p in prm_list:
            if p.startswith('*'): return aprms
            try: prms[p] = aprms[p]
            except(KeyError): pass
        return prms
    def set_params(self, prms):
        """Set all parameters from a dictionary."""
        changed = False
        self.beam.set_params(prms)
        try: self.pos[0], changed = prms['x'], True
        except(KeyError): pass
        try: self.pos[1], changed = prms['y'], True
        except(KeyError): pass
        try: self.pos[2], changed = prms['z'], True
        except(KeyError): pass
        try: self._phsoff[-2], changed = prms['dly'], True
        except(KeyError): pass
        try: self._phsoff[-1], changed = prms['off'], True
        except(KeyError): pass
        try: self._phsoff, changed = prms['phsoff'], True
        except(KeyError): pass
        try: self.bp_r, changed = prms['bp_r'], True
        except(KeyError): pass
        try: self.bp_i, changed = prms['bp_i'], True
        except(KeyError): pass
        try: self.amp, changed = prms['amp'], True
        except(KeyError): pass
        if changed: self.update()
        return changed

#     _          _                            _                         
#    / \   _ __ | |_ ___ _ __  _ __   __ _   / \   _ __ _ __ __ _ _   _ 
#   / _ \ | '_ \| __/ _ \ '_ \| '_ \ / _` | / _ \ | '__| '__/ _` | | | |
#  / ___ \| | | | ||  __/ | | | | | | (_| |/ ___ \| |  | | | (_| | |_| |
# /_/   \_\_| |_|\__\___|_| |_|_| |_|\__,_/_/   \_\_|  |_|  \__,_|\__, |
#                                                                 |___/ 

class AntennaArray(amp.AntennaArray):
    """Representation of location and time of observation, and response of
    array of antennas as function of pointing and frequency.  Adds get_params()
    and set_params() to amp.AntennaArray."""
    def get_params(self, ant_prms={'*':'*'}):
        """Return all fitable parameters in a dictionary."""
        prms = {}
        for k in ant_prms:
            if k.startswith('*'): ants = map(str, range(len(self)))
            else: ants = [k]
            prm_list = ant_prms[k]
            if type(prm_list) is str: prm_list = [prm_list]
            for a in ants:
                try: prms[a] = self.ants[int(a)].get_params(prm_list)
                except(ValueError): pass
        return prms
    def set_params(self, prms):
        """Set all parameters from a dictionary."""
        changed = False
        for i, a in enumerate(self):
            try: changed |= a.set_params(prms[str(i)])
            except(KeyError): pass
        if changed: self.update()
        return changed



#!/usr/global/paper/bin/python
"""
Script for displaying a projection of a spherical (Healpix) data set stored
in a *.fits file.

Author: Aaron Parsons
"""

import aipy as a, numpy as np, sys, os, ephem, optparse
from matplotlib import pylab as p

class Basemap:
    """A placeholder class to give plot_map.py some functionality if
    matplotlib-basemap is not installed."""
    def __init__(self, projection, lat_0=0, lon_0=0, **kwargs):
        if projection != 'cyl':
            raise ValueError('Without matplotlib-basemap installed, only cyl projection is supported.')
        self.lat_0, self.lon_0 = lat_0, lon_0
    def __call__(self, lon, lat, inverse=False):
        if inverse:
            lon = lon.astype(np.float) * 90 - self.lon_0
            lon = self.wrap(lon, -180, 180)
            lat = lat.astype(np.float) * 90 - self.lat_0
            lat = self.wrap(lat, -90, 90)
            return lon,lat
        else:
            x = (lon + self.lon_0) / 90.
            x = self.wrap(x, -2, 2)
            y = (lat + self.lat_0) / 90.
            y = self.wrap(y, -1, 1)
            return x,y
    def wrap(self, data, lo, hi):
        data = np.where(data >= hi, lo + (data - hi), data)
        data = np.where(data < lo, hi + (data - lo), data)
        return data
    def drawmapboundary(self): pass
    def drawmeridians(self, locs, **kwargs):
        x,y = self(locs, np.zeros_like(locs))
        x = self.wrap(x, -2, 2)
        p.xticks(x, visible=False)
        p.grid(True)
    def drawparallels(self, lats, **kwargs):
        x,y = self(np.zeros_like(lats),lats)
        y = self.wrap(y, -1, 1)
        p.yticks(y, [str(L) for L in lats])
        p.grid(True)
    def makegrid(self, dim1, dim2, returnxy=True):
        y,x = np.indices((dim2,dim1))
        x = 4 * x.astype(np.float)/dim1 - 2
        y = 1 - 2 * y.astype(np.float)/dim2
        lon,lat = self(x,y, inverse=True)
        if returnxy: return lon,lat, x,y
        else: return lon,lat
    def imshow(self, *args, **kwargs):
        kwargs['extent'] = (-2,2,-1,1)
        return p.imshow(*args, **kwargs)

# Try to import basemap module, but on failure use the above class
try: from mpl_toolkits.basemap import Basemap
except(ImportError): 
    try: from matplotlib.toolkits.basemap import Basemap
    except(ImportError): pass

o = optparse.OptionParser()
o.set_usage('plot_map.py [options] mapfile')
o.set_description(__doc__)
a.scripting.add_standard_options(o, cal=True, src=True, 
    cmap=True, max=True, drng=True)
o.add_option('-p', '--projection', dest='projection', default='moll',
    help='Map projection to use: moll (default), mill, cyl, robin, sinu.')
o.add_option('-m', '--mode', dest='mode', default='log',
    help='Plotting mode, can be log (default), lin.')
o.add_option('--interpolation', dest='interpolation', default='nearest',
    help='Sub-pixel interpolation.  Can be "nearest" or "bicubic".  Default nearest.')
o.add_option('-c', '--cen', dest='cen', type='float', 
    help="Center longitude/right ascension (in degrees) of map.  Default is 0 for galactic coordinate output, 180 for equatorial.")
o.add_option('-j', '--juldate', dest='juldate', type='float', 
    help='Julian date used for locating moving sources.')
o.add_option('--src_mark', dest='src_mark', default='',
    help='Marker to put on src locations.  Can be: ".,o,+,x,^,v".  Default no marker.')
o.add_option('--src_color', dest='src_color', default='k',
    help='Color of source label.  Can be: "k,w,r,b".  Default "k".')
o.add_option('-o', '--outfile', dest='outfile', default='',
    help='If provided, will save the figure to the specified file instead of popping up a window.')
o.add_option('--isys', dest='isys', default='eq',
    help='Input coordinate system (in map).  Can be eq (equatorial, default), ga (galactic), or ec (ecliptic).')
o.add_option('--osys', dest='osys', default='eq',
    help='Output coordinate system (plotted).  Can be eq (equatorial, default), ga (galactic), or ec (ecliptic)')
o.add_option('--iepoch', dest='iepoch', type='float', default=ephem.J2000,
    help='Epoch of input coordinates (in map).  Default J2000.')
o.add_option('--oepoch', dest='oepoch', type='float', default=ephem.J2000,
    help='Epoch of output coordinates (plotted).  Default J2000.')
o.add_option('--nobar', dest='nobar', action='store_true',
    help="Do not show colorbar.")
o.add_option('--res', dest='res', type='float', default=0.25,
    help="Resolution of plot (in degrees).  Default 0.25.")
o.add_option('--nside', dest='nside', type='int',
    help="Manually set NSIDE (possibly degrading map) to a power of 2.")
o.add_option('--mask', dest='mask', type='float',
    help="Optional dB of weight below which data will be masked. Recommended=30")
opts,args = o.parse_args(sys.argv[1:])

cmap = p.get_cmap(opts.cmap)
if opts.cen is None:
    if opts.osys == 'eq': opts.cen = 180
    else: opts.cen = 0
map = Basemap(projection=opts.projection,lat_0=0,lon_0=opts.cen, rsphere=1.)
lons,lats,x,y = map.makegrid(360/opts.res,180/opts.res, returnxy=True)
# Mask off parts of the image to be plotted that are outside of the map
lt = lats[:,0]
ln1 = np.ones_like(lt) * (lons[lons.shape[0]/2,0])
ln2 = np.ones_like(lt) * (lons[lons.shape[0]/2,-1])
x1,y1 = map(ln1,lt); x2,y2 = map(ln2,lt)
x = np.ma.array(x)
for c,(i,j) in enumerate(zip(x1,x2)): x[c] = np.ma.masked_outside(x[c], i, j)
mask = x.mask
if opts.osys == 'eq': lons = 360 - lons
lats *= a.img.deg2rad; lons *= a.img.deg2rad
print 'Reading %s' % args[0]
h = a.map.Map(fromfits=args[0])
print 'SCHEME:', h.scheme()
print 'NSIDE:', h.nside()
if not opts.nside is None:
    nh = a.healpix.HealpixMap(nside=opts.nside)
    nh.from_hpm(h)
    h = nh
h.set_interpol(opts.interpolation != 'nearest')

if opts.osys == 'eq': crd = a.coord.radec2eq(np.array([lons.flatten(), lats.flatten()]))
else: crd = a.coord.radec2eq(np.array([-lons.flatten(), lats.flatten()]))
m = a.coord.convert_m(opts.osys, opts.isys, 
    iepoch=opts.oepoch, oepoch=opts.iepoch)
x,y,z = np.dot(m, crd)
try: data, indices = h[x,y,z]
except(ValueError): data = h[x,y,z]
if not opts.mask is None:
    try:
        wgts = h.wgt[x,y,z]
        threshold = 10**(-opts.mask/10.)*np.max(wgts)
        msk = np.where(wgts > threshold, 1, 0)
        data *= msk
        print "Masking %2.0f%% of sky"% ((1 - msk.sum() / float(len(msk)))*100)
    except(AttributeError):
        print "Weights not included in file. No mask will be applied."
data.shape = lats.shape


# Generate source locations
if not opts.src is None:
    srclist,cutoff,catalogs = a.scripting.parse_srcs(opts.src, opts.cat)
    if not opts.cal is None:
        cat = a.cal.get_catalog(opts.cal, srclist, cutoff, catalogs)
    else:
        cat = a.src.get_catalog(srclist, cutoff, catalogs)
    o = ephem.Observer()
    if opts.juldate is None:
        o.date = ephem.J2000
        o.epoch = o.date
        try: del(cat['Sun'])
        except(KeyError): pass
    else:
        o.date = a.phs.juldate2ephem(opts.juldate)
        o.epoch = o.date
    for s in cat.values():
        try: a.phs.RadioFixedBody.compute(s, o)
        except(TypeError): a.phs.RadioSpecial.compute(s, o)
    #cat.compute(o)
    # lat/lon coordinates of sources
    scrds = [ephem.Equatorial(s.ra,s.dec,epoch=o.epoch) for s in cat.values()]
    afreqs = np.array([.150])
    cat.update_jys(afreqs)
    sflxs = cat.get_jys().squeeze()
    snams = cat.keys()
    if opts.osys == 'ga':
        scrds = [ephem.Galactic(s, epoch=opts.oepoch) for s in scrds]
    elif opts.osys == 'ec':
        scrds = [ephem.Ecliptic(s, epoch=opts.oepoch) for s in scrds]
    slats = np.array([float(s.get()[1]) for s in scrds]) * a.img.rad2deg
    slons = np.array([float(s.get()[0]) for s in scrds]) * a.img.rad2deg
    if opts.osys == 'eq': slons = 360 - slons
    slons = np.where(slons < -180, slons + 360, slons)
    slons = np.where(slons >= 180, slons - 360, slons)

# Generate map grid/outline
map.drawmapboundary()
map.drawmeridians(np.arange(-180, 180, 30))
map.drawparallels(np.arange(-90,90,30)[1:], labels=[0,1,0,0], labelstyle='+/-')
# Set up data to plot
if opts.mode.startswith('log'): data = np.log10(np.abs(data))
if opts.max is None: max = data.max()
else: max = opts.max
if opts.drng is None:
    min = data.min()
    if min < (max - 10): min = max-10
else: min = max - opts.drng
data = data.clip(min, max)
data = np.ma.array(data, mask=mask)
map.imshow(data, vmax=max, vmin=min, cmap=cmap, interpolation=opts.interpolation)

# Plot src labels and markers on top of map image
if not opts.src is None:
    sx, sy = map(slons,slats)
    for name, xpt, ypt, flx in zip(snams, sx, sy, sflxs):
        if xpt >= 1e30 or ypt >= 1e30: continue
        if opts.src_mark != '':
            map.plot(sx, sy, opts.src_color+opts.src_mark,markerfacecolor=None)
        if flx < 10: flx = 10
        p.text(xpt+.001, ypt+.001, name, size=5+2*int(np.round(np.log10(flx))),
            color=opts.src_color)
if not opts.nobar: p.colorbar(shrink=.5, format='%.2f')
else: p.subplots_adjust(.05,.05,.95,.95)


def mk_arr(val, dtype=np.double):
    if type(val) is np.ndarray: return val.astype(dtype)
    return np.array(val, dtype=dtype).flatten()

if opts.outfile != '':
    print 'Saving to', opts.outfile
    p.savefig(opts.outfile)
else:
    # Add right-click functionality for finding locations/strengths in map.
    cnt = 1
    def click(event):
        global cnt
        if event.button == 3: 
            lon,lat = map(event.xdata, event.ydata, inverse=True)
            if opts.osys == 'eq': lon = (360 - lon) % 360
            lon *= a.img.deg2rad; lat *= a.img.deg2rad
            ra,dec = ephem.hours(lon), ephem.degrees(lat)
            x,y,z = a.coord.radec2eq((ra,dec))
            flx = h[(x,y,z)]
            print '#%d (RA,DEC): (%s, %s), Jy: %f' % (cnt, ra, dec, flx)
            cnt += 1
        elif event.button==2:
            lon,lat = map(event.xdata, event.ydata, inverse=True)
            if opts.osys == 'eq': lon = (360 - lon) % 360
            lon *= a.img.deg2rad; lat *= a.img.deg2rad
            ra,dec = ephem.hours(lon), ephem.degrees(lat)
            x,y,z = a.coord.radec2eq((ra,dec))
            #flx = h[(x,y,z)]
            crd = [mk_arr(c, dtype=np.double) for c in (x,y,z)]
            px,wgts = h.crd2px(*crd, **{'interpolate':1})
            flx = np.sum(h[px],axis=-1)
            print '#%d (RA,DEC): (%s, %s), Jy: %f (4px sum)' % (cnt, ra, dec, flx)
            cnt += 1
        else: return
            


    #register this function with the event handler
    p.connect('button_press_event', click)
    p.show()



# -*- coding: utf-8 -*-
import unittest, ephem, random
import aipy as a, numpy as n

class TestPointingError(unittest.TestCase):
    def setUp(self):
        self.pnterr = a.phs.PointingError('Error String')
    def teststring(self):
        """Test setting a PointingError message string"""
        self.assertEqual(str(self.pnterr), 'Error String')
    def testraise(self):
        """Test raising a PointingError"""
        def raise_pnterr(): raise self.pnterr
        self.assertRaises(a.phs.PointingError, raise_pnterr)

class TestJulDates(unittest.TestCase):
    def setUp(self):
        self.ephemzero = ephem.date('1899/12/31 12:00')
        self.jdzero = 2415020.
    def testephemzero(self):
        """Test converting between ephem dates and JD - zero point"""
        self.assertEqual(a.phs.juldate2ephem(self.jdzero), self.ephemzero)
        self.assertEqual(a.phs.ephem2juldate(self.ephemzero), self.jdzero)
    def testrandom(self):
        """Test converting between ephem dates and JD - various"""
        for i in range(10):
            d1 = random.random() * ephem.now()
            d2 = a.phs.juldate2ephem(a.phs.ephem2juldate(d1))
            self.assertAlmostEqual(d1, d2)

class TestRadioBody(unittest.TestCase):
    def test_attributes(self):
        """Test aipy.phs.RadioFixedBody attributes"""
        epoch = ephem.B1950
        s = a.phs.RadioFixedBody('0:00', '0:00', mfreq=.200, name='src1',
            epoch=epoch, ionref=(.0,.0), srcshape=(.003, .005, .6))
        self.assertEqual(s._ra, ephem.hours('0:00'))
        self.assertEqual(s._dec, ephem.degrees('0:00'))
        self.assertEqual(s.mfreq, .200)
        self.assertEqual(s.src_name, 'src1')
        self.assertAlmostEqual(s._epoch, epoch, 3)
        self.assertEqual(s.ionref, [.0, .0])
        self.assertEqual(s.srcshape, [.003, .005, .6])
        self.assertRaises(RuntimeError, lambda: s.ra)
        self.assertRaises(RuntimeError, lambda: s.dec)
        self.assertRaises(AttributeError, lambda: s.map)
        o = a.phs.ArrayLocation(('0:00','0:00'))
        o.set_ephemtime(epoch)
        s.compute(o)
        self.assertEqual(len(s.get_crds('eq', ncrd=2)), 2)
        self.assertEqual(len(s.get_crds('top', ncrd=2)), 2)
        self.assertEqual(len(s.get_crds('eq', ncrd=3)), 3)
        self.assertEqual(len(s.get_crds('top', ncrd=3)), 3)
        self.assertEqual(s.map.shape, (3,3))
    def testephem(self):
        """Test the aipy.phs.RadioFixedBody ephem interface"""
        o = a.phs.ArrayLocation(('0:00','0:00'))
        for epoch in [ephem.B1900, ephem.B1950, ephem.J2000]:
          for ra in n.arange(1*n.pi/8, 2*n.pi, n.pi/8):
            for dec in n.arange(-3*n.pi/8, n.pi/2, n.pi/8):
                s = a.phs.RadioFixedBody(ra, dec, epoch=epoch)
                o.set_ephemtime(epoch)
                s.compute(o)
                self.assertAlmostEqual(s.a_ra, s._ra, 9)
                self.assertAlmostEqual(s.a_dec, s._dec, 9)
    def test_compute(self):
        """Test the aipy.phs.RadioFixedBody ephem calculations"""
        epoch = ephem.J2000
        o = a.phs.ArrayLocation(('0:00','0:00'))
        o.set_ephemtime(epoch)
        diagonal = n.array([[1.,0,0],[0,1,0],[0,0,1]])
        for ra in n.arange(1*n.pi/8, 2*n.pi, n.pi/8):
          for dec in n.arange(-3*n.pi/8, n.pi/2, n.pi/8):
            s = a.phs.RadioFixedBody(ra, dec, epoch=epoch)
            s.compute(o)
            m = a.coord.top2eq_m(o.sidereal_time()-s.ra, s.dec)
            err = n.abs(diagonal  - n.dot(m, s.map)).sum()
            self.assertAlmostEqual(err, 0, 10)
    def test_get_crds(self):
        """Test the aipy.phs.RadioFixedBody calculated coordinates"""
        epoch = ephem.J2000
        o = a.phs.ArrayLocation(('0:00','0:00'))
        o.set_ephemtime(epoch)
        for ra in n.arange(1*n.pi/8, 2*n.pi, n.pi/8):
          for dec in n.arange(-3*n.pi/8, n.pi/2, n.pi/8):
            s = a.phs.RadioFixedBody(ra, dec, epoch=epoch)
            s.compute(o)
            ra, dec = s.get_crds('eq', ncrd=2)
            self.assertEqual(s.ra, ra); self.assertEqual(s.dec, dec)
            az, alt = s.get_crds('top', ncrd=2)
            self.assertEqual(s.az, az); self.assertEqual(s.alt, alt)
            eq = s.get_crds('eq', ncrd=3)
            ra, dec = a.coord.eq2radec(eq)
            self.assertAlmostEqual(s.ra, ra, 10)
            self.assertAlmostEqual(s.dec, dec, 10)
            top = s.get_crds('top', ncrd=3)
            az, alt = a.coord.top2azalt(top)
            self.assertAlmostEqual(s.az, az, 10)
            self.assertAlmostEqual(s.alt, alt, 10)

class TestSrcCatalog(unittest.TestCase):
    def setUp(self):
        src1 = a.phs.RadioFixedBody('1:00', '1:00', name='src1')
        src2 = a.phs.RadioFixedBody('2:00', '2:00', name='src2')
        src3 = a.phs.RadioFixedBody('3:00', '3:00', name='src3')
        self.srcs = [src1, src2, src3]
        self.cat = a.phs.SrcCatalog(self.srcs)
    def test_add_srcs(self):
        """Test adding sources to a aipy.phs.SrcCatalog() catalog"""
        cat = a.phs.SrcCatalog()
        src1b = a.phs.RadioFixedBody('0:00', '0:00', name='src1')
        src4 = a.phs.RadioFixedBody('4:00', '4:00', name='src4')
        cat.add_srcs(self.srcs)
        self.assertEqual(len(cat), 3)
        cat.add_srcs(src1b)
        self.assertEqual(len(cat), 3)
        cat.add_srcs(src1b, src4)
        self.assertEqual(len(cat), 4)
        srclist = [src1b] + self.srcs[1:] + [src4]
        for name, src in zip([s.src_name for s in srclist], srclist):
            self.assertEqual(src, cat[name])
    def test_get_srcs(self):
        """Test retrieving sources from a aipy.phs.SrcCatalog() catalog"""
        self.assertEqual(self.cat.get_srcs('src1','src2','src3'), self.srcs)
        self.assertEqual(self.cat.get_srcs(['src1','src2']), self.srcs[:2])
        self.assertRaises(KeyError, lambda: self.cat.get_srcs('bad'))
    def test_compute(self):
        """Test the ephem interfaces for a aipy.phs.SrcCatalog() catalog"""
        o = ephem.Observer()
        self.cat.compute(o)
        for src in self.cat.values():
            self.assertNotEqual(src.ra, None)
            self.assertNotEqual(src.dec, None)
    def test_get_crds(self):
        """Test coordinates calculated from a aipy.phs.SrcCatalog() catalog"""
        o = ephem.Observer()
        self.cat.compute(o)
        crd1 = self.cat.get_crds('eq', srcs=['src1'])
        self.assertEqual(crd1.shape, (3,1))
        self.assertTrue(n.all(crd1[:,0] == self.srcs[0].get_crds('eq')))
        crd2 = self.cat.get_crds('top', srcs=['src1','src2'])
        self.assertEqual(crd2.shape, (3,2))
    def test_get(self):
        """Test retrieving source attributes from a aipy.phs.SrcCatalog() catalog"""
        mfreq = self.cat.get('mfreq',srcs=['src1'])
        self.assertEqual(mfreq.shape, (1,))
        mfreq = self.cat.get('mfreq',srcs=['src1','src2'])
        self.assertEqual(mfreq.shape, (2,))
        ionrefs = self.cat.get('ionref',srcs=['src1','src2'])
        self.assertEqual(ionrefs.shape, (2,2))
        srcshapes = self.cat.get('srcshape',srcs=['src1','src2'])
        self.assertEqual(srcshapes.shape, (3,2))

class TestBeam(unittest.TestCase):
    def setUp(self):
        self.fq = n.arange(0,1,.1)
        self.bm = a.phs.Beam(self.fq)
    def test_attributes(self):
        """Test accessing aipy.phs.Beam attributes"""
        self.assertTrue(n.all(self.bm.freqs == self.fq))
        self.assertTrue(n.all(self.bm.chans == n.arange(self.fq.size)))
        self.assertTrue(n.all(self.bm.afreqs == self.fq))
    def test_select_chans(self):
        """Test selecting various aipy.phs.Beam channels"""
        chans = n.array([1,2,3])
        self.bm.select_chans(chans)
        self.assertTrue(n.all(self.bm.chans == chans))
        self.assertTrue(n.all(self.bm.afreqs == self.fq.take(chans)))

class TestAntenna(unittest.TestCase):
    def setUp(self):
        fq = n.arange(0,1,.1)
        self.bm = a.phs.Beam(fq)
        self.ant = a.phs.Antenna(1, 2, 3, self.bm, phsoff=[0,1])
    def test_attributes(self):
        """Test accessing aipy.phs.Antenna attributes"""
        self.assertEqual(self.ant.beam, self.bm)
        pos = n.array([1,2,3], n.float64)
        self.assertTrue(n.all(self.ant.pos == pos))
        x,y,z = self.ant
        self.assertEqual(x, pos[0])
        self.assertEqual(y, pos[1])
        self.assertEqual(z, pos[2])
        self.assertTrue(n.all(self.ant + self.ant == pos*2))
        self.assertTrue(n.all(self.ant - self.ant == 0))
        self.assertTrue(n.all(-self.ant == -pos))
    def test_select_chans(self):
        """Test selecting various aipy.phs.Antenna channels"""
        chans = n.array([1,2,3])
        self.ant.select_chans(chans)
        self.assertTrue(n.all(self.bm.chans == chans))
        self.assertTrue(n.all(self.ant.phsoff == 1))
        
class TestArrayLocation(unittest.TestCase):
    def setUp(self):
        self.aa = a.phs.ArrayLocation(('0','0'))
    def test_attributes(self):
        self.assertEqual(self.aa.pressure, 0)
        self.assertEqual(self.aa.lat, 0)
        self.assertEqual(self.aa.long, 0)
        self.assertEqual(self.aa.elev, 0)
        self.assertTrue(n.all(self.aa._eq2zen == a.coord.eq2top_m(0., 0.)))
    def test_set_jultime(self):
        jd = 2454554.9798841
        self.aa.set_jultime(jd)
        self.assertEqual(self.aa.date, self.aa.epoch)
        self.assertEqual(self.aa.date, a.phs.juldate2ephem(jd))
        self.assertAlmostEqual(self.aa.sidereal_time(), 0, 7)
        eq2now_rnd = n.round(self.aa._eq2now, 7)
        self.assertTrue(n.all(eq2now_rnd ==
            n.array([[1.,0.,0.], [0.,1.,0.], [0.,0.,1.]])))
    def test_get_jultime(self):
        self.aa.set_jultime(2454555)
        self.assertEqual(self.aa.get_jultime(), 2454555)

class TestAntennaArray(unittest.TestCase):
    def setUp(self):
        bm = a.phs.Beam(n.arange(0,1,.1))
        a1 = a.phs.Antenna(0,0,0,bm, [0,0])
        a2 = a.phs.Antenna(1,0,0,bm, [0,1])
        a3 = a.phs.Antenna(0,1,0,bm, [0,2])
        a4 = a.phs.Antenna(0,0,1,bm, [0,3])
        self.ants = [a1,a2,a3,a4]
        self.aa = a.phs.AntennaArray(('0','0'), self.ants)
    def test_attributes(self):
        self.assertEqual(len(self.aa), 4)
        for ai, aj in zip(self.aa, self.ants): self.assertEqual(ai, aj)
        for i in range(4): self.assertEqual(len(self.aa[:i]), i)
    def test_select_chans(self):
        chans = n.array([1,2,3])
        self.aa.select_chans(chans)
        for ant in self.aa.ants:
            self.assertTrue(n.all(ant.beam.chans == chans))
    def test_ij2bl(self):
        self.assertEqual(self.aa.ij2bl(0,1), 258)
        self.assertEqual(self.aa.ij2bl(0,2), 259)
        self.assertEqual(self.aa.ij2bl(1,2), 515)
    def test_bl2ij(self):
        self.assertEqual(self.aa.bl2ij(258), (0,1))
        self.assertEqual(self.aa.bl2ij(259), (0,2))
        self.assertEqual(self.aa.bl2ij(515), (1,2))
    def test_get_baseline(self):
        for j, ant in enumerate(self.aa):
            if j in [0,3]:
                self.aa.set_jultime(2454554.9)
                self.assertTrue(n.all(self.aa.get_baseline(0,j,'r') == ant.pos))
                self.assertTrue(n.all(self.aa.get_baseline(0,j,'e') == ant.pos))
            else:
                self.assertTrue(n.all(self.aa.get_baseline(0,j,'r') == ant.pos))
                self.aa.set_jultime(2454554.9798841)
                bl_rnd = n.round(self.aa.get_baseline(0,j,'e'), 7)
                self.assertTrue(n.all(bl_rnd == ant.pos))
                self.aa.set_jultime(2454554.9)
                bl_rnd = n.round(self.aa.get_baseline(0,j,'e'), 7)
                self.assertFalse(n.all(bl_rnd == ant.pos))
        src = a.phs.RadioFixedBody('12:00', '0:00')
        src.compute(self.aa)
        self.assertRaises(a.phs.PointingError, 
            lambda: self.aa.get_baseline(0,1,src))
        for t in n.random.random((10,)):
            self.aa.set_jultime(2454555. + t)
            src = a.phs.RadioFixedBody(self.aa.sidereal_time(), self.aa.lat,
                epoch=self.aa.epoch)
            src.compute(self.aa)
            zbl_rnd = n.round(self.aa.get_baseline(0,1,'z'), 3)
            sbl_rnd = n.round(self.aa.get_baseline(0,1,src), 3)
            self.assertTrue(n.all(zbl_rnd == sbl_rnd))
    def test_get_phs_offset(self):
        self.assertTrue(n.all(self.aa.get_phs_offset(0,0) == 0))
        self.assertTrue(n.all(self.aa.get_phs_offset(0,1) == 1))
        self.assertTrue(n.all(self.aa.get_phs_offset(0,2) == 2))
        self.assertTrue(n.all(self.aa.get_phs_offset(0,3) == 3))
    def test_gen_uvw(self):
        self.aa.select_chans()
        afreqs = self.aa[0].beam.afreqs
        u,v,w = self.aa.gen_uvw(0,1,'z')
        self.assertEqual(u.shape, (1,afreqs.size))
        self.assertTrue(n.all(u == 0*afreqs))
        self.assertTrue(n.all(v == 0*afreqs))
        self.assertTrue(n.all(w == 1*afreqs))
        u,v,w = self.aa.gen_uvw(0,2,'z')
        self.assertTrue(n.all(u == 1*afreqs))
        self.assertTrue(n.all(v == 0*afreqs))
        self.assertTrue(n.all(w == 0*afreqs))
        u,v,w = self.aa.gen_uvw(0,3,'z')
        self.assertTrue(n.all(u == 0*afreqs))
        self.assertTrue(n.all(v == 1*afreqs))
        self.assertTrue(n.all(w == 0*afreqs))
    def test_gen_phs(self):
        self.aa.select_chans([1,2,3])
        afreqs = self.aa[0].beam.afreqs
        for t in n.random.random((40,)):
            self.aa.set_jultime(2454555. + t)
            src = a.phs.RadioFixedBody(self.aa.sidereal_time(), self.aa.lat,
                epoch=self.aa.epoch)
            src.compute(self.aa)
            if t > .5: 
                seq = src.get_crds('eq', ncrd=3)
                if t > .75: seq = n.array([seq,seq]).transpose()
            else: seq = src
            phs = n.round(self.aa.gen_phs(seq, 0, 1, mfreq=.1), 6)
            ans = n.round(n.exp(-1j*2*n.pi*afreqs), 6)
            if t > .75: self.assertEqual(phs.shape, (2,3))
            else: self.assertEqual(phs.shape, (3,))
            self.assertTrue(n.all(phs == ans))
            phs = n.round(self.aa.gen_phs(src, 0, 2, mfreq=.1), 3)
            self.assertTrue(n.all(phs == 1+0j))
            phs = n.round(self.aa.gen_phs(src, 0, 3, mfreq=.1), 3)
            self.assertTrue(n.all(phs == 1+0j))
        phs1 = self.aa.gen_phs(src, 0, 2, mfreq=.1, ionref=(.001,.001))
        phs2 = self.aa.gen_phs(src, 0, 2, mfreq=.1, srcshape=(.01,.01,0), 
            resolve_src=True)
        self.assertTrue(n.all(phs1 != 1+0j))
        self.assertTrue(n.all(phs2 != 1+0j))
    def test_resolve_src(self):
        amp = self.aa.resolve_src(100., 100., srcshape=(0,0,0))
        self.assertEqual(amp, 1)
        amp1 = self.aa.resolve_src(100., 50., srcshape=(.01,0,n.pi/2))
        amp2 = self.aa.resolve_src(100., 50., srcshape=(0,.01,0))
        self.assertAlmostEqual(amp1, amp2, 15)
        amp1 = self.aa.resolve_src(100., 100., srcshape=(.02,0,0))
        amp2 = self.aa.resolve_src(100., 100., srcshape=(.01414,.01414,0))
        self.assertAlmostEqual(amp1, amp2, 3)
        amp = self.aa.resolve_src(100., 0., srcshape=(0.001,0,0))
        x = 2*n.pi * .1
        self.assertEqual(amp, 2*a._cephes.j1(x)/x)
    def test_refract(self):
        self.aa.select_chans([1,2,3])
        afreqs = self.aa[0].beam.afreqs
        zeros = n.zeros((1,3), dtype=n.float)
        ones = n.ones((1,3), dtype=n.float)
        # Test non-vectors, dra->u association
        dw = self.aa.refract(ones, zeros, mfreq=.1, ionref=(.001,0))
        ans = .001 / (afreqs/.1)**2 ; ans.shape = (1, ans.size)
        self.assertEqual(len(dw.shape), 2)
        self.assertTrue(n.all(n.round(dw,10) == n.round(ans,10)))
        # Test non-vectors, no dra-> v association
        dw = self.aa.refract(zeros, ones, mfreq=.1,ionref=(.001,0))
        self.assertTrue(n.all(dw == 0))
        # Test non-vectors, no ddec->u association
        dw = self.aa.refract(ones, zeros, mfreq=.1, ionref=(0,.001))
        self.assertTrue(n.all(dw == 0))
        # Test non-vectors, ddec->v association, v scaling
        dw = self.aa.refract(zeros, 2*ones, mfreq=.1, ionref=(0,.001))
        self.assertTrue(n.all(n.round(dw,10) == n.round(2*ans,10)))
        # Test vectors, mfreq scaling
        ones = n.ones((2,3), dtype=n.float)
        ionref = (n.array([0, .001]), n.array([.001, 0]))
        mfreq = n.array([.1, .2])
        ans = n.array([.002/(afreqs / .1)**2, .002/(afreqs / .2)**2])
        dw = self.aa.refract(2*ones, 2*ones, mfreq=mfreq, ionref=ionref)
        self.assertTrue(n.all(n.round(dw,10) == n.round(ans,10)))
    def test_phs2src(self):
        self.aa.select_chans([1,2,3])
        self.aa.set_jultime(2454555.)
        src = a.phs.RadioFixedBody('0:00', '20:00')
        src.compute(self.aa)
        self.assertTrue(n.all(self.aa.phs2src(1.,src,0,1) == \
            self.aa.gen_phs(src,0,1)))
    def test_unphs2src(self):
        self.aa.select_chans([1,2,3])
        self.aa.set_jultime(2454555.)
        src = a.phs.RadioFixedBody('0:00', '20:00')
        src.compute(self.aa)
        self.assertTrue(n.all(
            self.aa.unphs2src(self.aa.gen_phs(src,0,1),src,0,1) == 1.))

class TestSuite(unittest.TestSuite):
    """A unittest.TestSuite class which contains all of the aipy.phs unit tests."""

    def __init__(self):
        unittest.TestSuite.__init__(self)

        loader = unittest.TestLoader()
        self.addTests(loader.loadTestsFromTestCase(TestPointingError))
        self.addTests(loader.loadTestsFromTestCase(TestJulDates))
        self.addTests(loader.loadTestsFromTestCase(TestRadioBody))
        self.addTests(loader.loadTestsFromTestCase(TestSrcCatalog))
        self.addTests(loader.loadTestsFromTestCase(TestBeam))
        self.addTests(loader.loadTestsFromTestCase(TestAntenna))
        self.addTests(loader.loadTestsFromTestCase(TestArrayLocation))
        self.addTests(loader.loadTestsFromTestCase(TestAntennaArray))

if __name__ == '__main__':
    unittest.main()



#! /usr/bin/env python

from __future__ import absolute_import, division, print_function

from setuptools import setup, Extension

import os, glob, numpy, subprocess

print("Generating aipy_src/__version__.py: ", end='')
__version__ = open('VERSION').read().strip()
print(__version__)
open('aipy_src/__version__.py','w').write('__version__="%s"'%__version__)

#read the latest git status out to an installed file
try:
#    gitbranch = subprocess.check_output('git symbolic-ref -q HEAD',shell=True, cwd='.').strip().split('/')[-1]
    gitbranch = os.popen('git symbolic-ref -q HEAD').read().strip()
    print("Generating aipy_src/__branch__.py")
#    gitlog = subprocess.check_output('git log -n1 --pretty="%h%n%s%n--%n%an%n%ae%n%ai"',shell=True, cwd='.').strip()
    gitlog = os.popen('git log -n1 --pretty="%h%n%s%n--%n%an%n%ae%n%ai"').read().strip()
    print("Generating aipy_src/__gitlog__.py.")
    print(gitlog)
except:
    gitbranch = "unknown branch"
    gitlog = "git log not found"
open('aipy_src/__branch__.py','w').write('__branch__ = \"%s\"'%gitbranch)
open('aipy_src/__gitlog__.py','w').write('__gitlog__ = \"\"\"%s\"\"\"'%gitlog)


def get_description():
    lines = [L.strip() for L in open('README.md').readlines()]
    d_start = None
    for cnt, L in enumerate(lines):
        if L.startswith('## Description'): d_start = cnt + 1
        elif not d_start is None:
            if len(L) == 0: return ' '.join(lines[d_start:cnt])
    raise RuntimeError('Bad README')

def indir(path, files):
    return [os.path.join(path, f) for f in files]

global_macros = [('NPY_NO_DEPRECATED_API', 'NPY_1_7_API_VERSION')]

setup(name = 'aipy',
    version = __version__,
    description = 'Astronomical Interferometry in PYthon',
    long_description = get_description(),
    license = 'GPL',
    author = 'Aaron Parsons',
    author_email = 'aparsons@berkeley.edu',
    url = 'http://github.com/HERA-Team/aipy',
    classifiers = [
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License (GPL)',
        'Topic :: Scientific/Engineering :: Astronomy',
    ],
    setup_requires = ['numpy>=1.2'],
    install_requires = ['pyephem>=3.7.3.2', 'astropy>=1.0', 'numpy>=1.2'],
    package_dir = {'aipy':'aipy_src', 'aipy.optimize':'aipy_src/optimize', 'aipy._src':'aipy_src/_src'},
    packages = ['aipy', 'aipy.optimize','aipy._src'],
    ext_modules = [
        Extension('aipy._healpix',
            ['aipy_src/_healpix/healpix_wrap.cpp', 
            'aipy_src/_healpix/cxx/Healpix_cxx/healpix_base.cc'],
            define_macros = global_macros,
            include_dirs = [numpy.get_include(), 'aipy_src/_healpix/cxx/cxxsupport',
                'aipy_src/_healpix/cxx/Healpix_cxx']),
        Extension('aipy._alm',
            ['aipy_src/_healpix/alm_wrap.cpp', 
            'aipy_src/_healpix/cxx/Healpix_cxx/alm_map_tools.cc',
            'aipy_src/_healpix/cxx/libfftpack/ls_fft.c',
            'aipy_src/_healpix/cxx/libfftpack/bluestein.c',
            'aipy_src/_healpix/cxx/libfftpack/fftpack.c',
            'aipy_src/_healpix/cxx/Healpix_cxx/healpix_map.cc',
            'aipy_src/_healpix/cxx/Healpix_cxx/healpix_base.cc'],
            define_macros = global_macros,
            include_dirs = [numpy.get_include(), 'aipy_src/_healpix/cxx/cxxsupport',
                'aipy_src/_healpix/cxx/Healpix_cxx']),
        Extension('aipy._miriad', ['aipy_src/_miriad/miriad_wrap.cpp'] + \
            indir('aipy_src/_miriad/mir', ['uvio.c','hio.c','pack.c','bug.c',
                'dio.c','headio.c','maskio.c']),
            define_macros = global_macros,
            include_dirs = [numpy.get_include(), 'aipy_src/_miriad', 
                'aipy_src/_miriad/mir']),
        Extension('aipy._deconv', ['aipy_src/_deconv/deconv.cpp'],
            define_macros = global_macros,
            include_dirs = [numpy.get_include()]),
        #Extension('aipy._img', ['aipy_src/_img/img.cpp'],
        #    include_dirs = [numpy.get_include()]),
        Extension('aipy._dsp', ['aipy_src/_dsp/dsp.c', 'aipy_src/_dsp/grid/grid.c'],
            define_macros = global_macros,
            include_dirs = [numpy.get_include(), 'aipy_src/_dsp', 'aipy_src/_dsp/grid']),
        Extension('aipy.utils', ['aipy_src/utils/utils.cpp'],
            define_macros = global_macros,
            include_dirs = [numpy.get_include()]),
        Extension('aipy._cephes',
            ['aipy_src/_cephes/_cephesmodule.c', 'aipy_src/_cephes/ufunc_extras.c'] + \
            glob.glob('aipy_src/_cephes/cephes/*.c') + \
            glob.glob('aipy_src/_cephes/c_misc/*.c'),
            define_macros = global_macros,
            include_dirs = [numpy.get_include()]),
    ],
    scripts=glob.glob('scripts/*'),
    package_data = {'aipy': ['_aipy_src/*.txt']},
)



#!/usr/bin/python
"""
Parse large data files by antenna number. Keep antenna 0 in all of them as a phase reference.
"""

import aipy as a
import sys,os,optparse

o = optparse.OptionParser()
o.set_usage('pull_ants.py [options] *.uv')
o.set_description(__doc__)
o.add_option('-p','--pols',dest='pols',type='str',default=8,help='Which polarizations do you like?')
opts,args = o.parse_args(sys.argv[1:])

for filename in args:

    outfile = filename+'P'
    print filename,'-->',outfile
    if os.path.exists(outfile):
        print 'File exists, skipping'
        continue

    pols2use = opts.pols.split(',')

    def mfunc(uv,p,d):
        pol = a.miriad.pol2str[uv['pol']] 
        if pol in pols2use: return p,d
        else: return None,None

    uvi = a.miriad.UV(filename)
    uvo = a.miriad.UV(outfile,status='new')
    uvo.init_from_uv(uvi)
    histstr = 'PULL POLS: pols='
    for pol in pols2use:
        histstr += str(pol)+','
        if pol == pols2use[-1]: histstr += '\n'
    uvo.pipe(uvi,mfunc=mfunc,append2hist=histstr)
    del uvo,uvi



#! /usr/bin/env python
"""
Apply calibration parameters to a data set.
"""

import aipy as a
import optparse, sys, os

o = optparse.OptionParser()
a.scripting.add_standard_options(o, cal=True)
o.add_option('--no_gain', action='store_true', help='Do not change the gain (i.e. no flux cal).')
o.add_option('--no_phs', action='store_true', help='Do not change the phase (i.e. no phase cal).')
opts,args = o.parse_args(sys.argv[1:])

uv = a.miriad.UV(args[0])
aa = a.cal.get_aa(opts.cal, uv['sdf'], uv['sfreq'], uv['nchan'])
del(uv)

def mfunc(uv, p, d, f):
    uvw,t,(i,j) = p
    pol = a.miriad.pol2str[uv['pol']]
    aa.set_active_pol(pol)
    if not opts.no_gain: d /= aa.passband(i, j)
    if not opts.no_phs: d = aa.phs2src(d, 'z', i, j)
    return p, d, f

for infile in args:
    outfile = infile+'C'
    print infile,'-->',outfile
    if os.path.exists(outfile):
        print 'File exists, skipping....'
        continue 

    uvi = a.miriad.UV(infile)
    uvo = a.miriad.UV(outfile, status='new')
    uvo.init_from_uv(uvi)
    uvo.pipe(uvi, mfunc, raw=True, append2hist='APPLY_CAL: ' + ' '.join(sys.argv) + '\n')



#! /usr/bin/env python
# -*- coding: utf-8 -*-

import unittest
import aipy.healpix as h, numpy as n

hpm = h.HealpixMap(32, 'RING')

c1 = n.ones((20,), dtype=n.double)
c2 = n.ones((20,), dtype=n.double)
c3 = n.ones((20,), dtype=n.double)

print hpm[0], hpm[0,0], hpm[0,0,0]
hpm[0] = 1
hpm[0,0] = 1
hpm[0,0,0] = 1
print hpm[0], hpm[0,0], hpm[0,0,0]

c = n.arange(15)
print hpm[c]
hpm[c] += c
print hpm[c]
hpm[c] = c
print hpm[c]

c1 = n.arange(-.5,.5,.1)
c2 = n.arange(-.5,.5,.1)
c3 = n.arange(-.5,.5,.1)

print hpm[c1,c2,c3]
hpm[c1,c2,c3] += 1
print hpm[c1,c2,c3]

hpm.set_map(n.arange(48, dtype=n.double))
print hpm.get_map()
hpm.change_scheme('NEST')
print hpm.get_map()
